from google.protobuf.json_format import MessageToDict
from ..proto import serialize_length_delimited, ComputeNodeFormat, parse_length_delimited
from ..node import Node
from .proto import (
    PermutiveWorkerConfiguration,
    SinkInput,
    ExportRole,
)

class DataSourcePermutive(Node):
    """
    Compute node that fetches a dataset from Permutive.
    """

    def __init__(
            self,
            name: str,
            credentials_dependency: str,
    ) -> None:
        config = PermutiveWorkerConfiguration(
            credentialsDependency=credentials_dependency,
            importRole={}
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.permutive-worker",
            dependencies=[credentials_dependency],
            output_format=ComputeNodeFormat.RAW
        )

class PermutiveDataTransformer(Node):
    """
    Compute node that transforms data into a Permutive compliant form.
    """

    def __init__(
            self,
            name: str,
            credentials_dependency: str,
            input: SinkInput,
            import_id: str,
            segment_name: str,
            segment_code: str,
            input_has_headers: bool,
    ) -> None:
        config = PermutiveWorkerConfiguration(
            credentialsDependency=credentials_dependency,
            exportRole=ExportRole(input=input),
            importId = import_id,
            segmentName = segment_name,
            segmentCode = segment_code,
            inputHasHeaders=input_has_headers,
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.permutive-worker",
            dependencies=[credentials_dependency, input.dependency],
            output_format=ComputeNodeFormat.RAW
        )

class PermutiveWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = PermutiveWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)

__all__ = [
    "DataSourcePermutive",
    "PermutiveDataTransformer",
    "PermutiveWorkerDecoder"
]



from .types import MatchingId, MatchingIdFormat, TableColumnHashingAlgorithm

# Map the user specified matching ID to the corresponding internal
# matching ID format and hashing algorithm.
MATCHING_ID_INTERNAL_LOOKUP = {
    MatchingId.STRING: (MatchingIdFormat.STRING, None),
    MatchingId.EMAIL: (MatchingIdFormat.EMAIL, None),
    MatchingId.HASHED_EMAIL: (
        MatchingIdFormat.EMAIL,
        TableColumnHashingAlgorithm.SHA256_HEX,
    ),
    MatchingId.PHONE_NUMBER: (MatchingIdFormat.PHONE_NUMBER_E164, None),
    MatchingId.HASHED_PHONE_NUMBER: (
        MatchingIdFormat.PHONE_NUMBER_E164,
        TableColumnHashingAlgorithm.SHA256_HEX,
    ),
}


import json
import struct

import asn1crypto.x509
import asn1crypto.pem
from sgx_ias_structs import QuoteBody as _QuoteBody
from oscrypto import asymmetric
from certvalidator import CertificateValidator, ValidationContext
from enum import IntFlag


from .types import IasResponse, Tcb, TcbInfoContainer, TcbLevel
from .proto import (
    AttestationSpecification, Fatquote, FatquoteEpid, FatquoteDcap,
    AttestationSpecificationIntelDcap
)
from ecdsa import VerifyingKey
from hashlib import sha256
from typing import List, Optional
import pem as Pem
from .certs import intel_sgx_ias_root_ca, intel_sgx_dcap_root_ca


class QuoteBody:
    """Represents the body of the SGX IAS Quote"""

    def __init__(self, quote):
        if not isinstance(quote, _QuoteBody):
            raise Exception("QuoteBody must be converted from sgx_ias_structs repr.")
        self.version = quote.version
        self.sign_type = quote.sign_type
        self.qe_svn = quote.qe_svn
        self.pce_svn = quote.pce_svn
        self.xeid = quote.xeid
        self.epid_group_id = quote.epid_group_id
        self.basename = quote.basename
        self.reportdata = quote.reportdata
        self.cpusvn = quote.cpusvn
        self.mrenclave = quote.mrenclave
        self.mrsigner = quote.mrsigner
        self.isvprodid = quote.isvprodid
        self.isvsvn = quote.isvsvn
        self.flags = quote.flags

    def __str__(self):
        obj_dict = {
            k: bytes(v).hex() if isinstance(v, (list, bytes)) else v
            for (k, v) in self.__dict__.items()
        }
        return json.dumps(obj_dict, indent=4)

    @classmethod
    def from_base64_string(cls, data):
        return cls(_QuoteBody.from_base64_string(data))


class CertificateValidationError(Exception):
    """Raised when the certificate validation failed"""

    pass


class SignatureValidationError(Exception):
    """Raised when the signature validation failed"""

    pass


class MessageParsingError(Exception):
    """Raised when the json message parsing failed"""

    pass


class ISVEnclaveQuoteStatusError(Exception):
    """Raised when the isv enclave quote status is not in an accepted status"""

    pass


class EnclaveQuoteFlagsError(Exception):
    """Raised when the flags in quote body don't match expected values"""

    pass


class MeasurementMismatchError(Exception):
    """Raised when enclave measurement does not match expected value"""

    pass


class Verification:
    class IASAttributeFlags(IntFlag):
        INIT = 1
        DEBUG = 2
        MODE64BIT = 4
        PROVISIONKEY = 8
        EINITTOKENKEY = 16

    def __init__(
            self,
            attestation_specification: AttestationSpecification,
    ):
        self.attestation_specification = attestation_specification
        self.check_known_root_ca = True

    # Warning: this is unsafe to do, only do this if you deliberately want to test fake roots (e.g. for mocked enclaves)
    def disable_known_root_ca_check(self):
        self.check_known_root_ca = False

    def _verify_epid(self, epid: FatquoteEpid) -> bytes:
        if self.attestation_specification.WhichOneof("attestation_specification") != "intelEpid":
            raise Exception(f'Incompatible attestation specification, expected EPID')
        spec_epid = self.attestation_specification.intelEpid
        trust_roots = [spec_epid.iasRootCaDer]
        validation_context = ValidationContext(trust_roots)
        validator = CertificateValidator(epid.iasCertificate, validation_context=validation_context)
        validator.validate_usage({"digital_signature"})

        # Step 2: verify message signature
        pubk = asymmetric.load_public_key(epid.iasCertificate)
        asymmetric.rsa_pkcs1v15_verify(pubk, epid.iasSignature, epid.iasResponseBody, "sha256")

        # Step 3: parse message and get quote body
        message_dic: IasResponse = json.loads(epid.iasResponseBody)
        quote_body_encoded = message_dic["isvEnclaveQuoteBody"]
        isv_enclave_quote_status = message_dic["isvEnclaveQuoteStatus"]
        quote_body = _QuoteBody.from_base64_string(quote_body_encoded)

        # Step 4: check isv enclave quote status
        if isv_enclave_quote_status == "OK":
            pass
        elif spec_epid.accept_configuration_needed and isv_enclave_quote_status == "CONFIGURATION_NEEDED":
            pass
        elif spec_epid.accept_group_out_of_date and isv_enclave_quote_status == "GROUP_OUT_OF_DATE":
            pass
        else:
            raise ISVEnclaveQuoteStatusError("Status " + isv_enclave_quote_status + " not accepted")

        # Step 5: check enclave quote flags
        flags = quote_body.flags
        if not (flags & Verification.IASAttributeFlags.INIT):
            raise EnclaveQuoteFlagsError
        if not (flags & Verification.IASAttributeFlags.MODE64BIT):
            raise EnclaveQuoteFlagsError
        if flags & Verification.IASAttributeFlags.DEBUG:
            if spec_epid.accept_debug:
                print("!!!WARNING!!! DEBUG quote is being accepted, quote is NOT to be trusted !!!WARNING!!!")
            else:
                raise EnclaveQuoteFlagsError

        # Step 6: assert expected measurement
        measurement = bytes(quote_body.mrenclave)
        if not (spec_epid.mrenclave == measurement):
            raise MeasurementMismatchError(
                f"Wrong measurement, expected measurement: {spec_epid.mrenclave.hex()}, actual measurement: {measurement.hex()}."
            )

        return quote_body.reportdata

    def _get_cpusvn(self, tcb: Tcb) -> List[int]:
        cpusvn_array = [
            tcb["sgxtcbcomp01svn"],
            tcb["sgxtcbcomp02svn"],
            tcb["sgxtcbcomp03svn"],
            tcb["sgxtcbcomp04svn"],
            tcb["sgxtcbcomp05svn"],
            tcb["sgxtcbcomp06svn"],
            tcb["sgxtcbcomp07svn"],
            tcb["sgxtcbcomp08svn"],
            tcb["sgxtcbcomp09svn"],
            tcb["sgxtcbcomp10svn"],
            tcb["sgxtcbcomp11svn"],
            tcb["sgxtcbcomp12svn"],
            tcb["sgxtcbcomp13svn"],
            tcb["sgxtcbcomp14svn"],
            tcb["sgxtcbcomp15svn"],
            tcb["sgxtcbcomp16svn"]
        ]
        return cpusvn_array

    def _dcap_get_cert(self, quote: bytes):
        auth_len = int.from_bytes(quote[1012:1014], "little")

        cert = quote[1014 + auth_len + 2 + 4:]
        return cert

    def _dcap_check_status(self, spec_dcap: AttestationSpecificationIntelDcap, status: str):
        if status == "UpToDate" or status == "SWHardeningNeeded":
            return
        if spec_dcap.accept_out_of_date and status in [
            "OutOfDate",
            "OutOfDateConfigurationNeeded",
        ]:
            return
        if spec_dcap.accept_configuration_needed and status in [
            "ConfigurationNeeded",
            "OutOfDateConfigurationNeeded",
            "ConfigurationAndSWHardeningNeeded",
        ]:
            return
        if spec_dcap.accept_revoked and status in [
            "Revoked"
        ]:
            return
        raise Exception(f'TCB status ${status} not accepted')

    def _dcap_find_tcb_level(
            self,
            tcb_levels: List[TcbLevel],
            cpusvn: List[int],
            pcesvn: int
    ) -> Optional[TcbLevel]:
        for level in tcb_levels:
            tcb_cpusvn = self._get_cpusvn(level["tcb"])
            if all([a >= b for a, b in zip(cpusvn, tcb_cpusvn)]):
                if pcesvn >= level["tcb"]["pcesvn"]:
                    return level
        return None

    def _verify_dcap(self, dcap: FatquoteDcap) -> bytes:
        if self.attestation_specification.WhichOneof("attestation_specification") != "intelDcap":
            raise Exception(f'Incompatible attestation specification, expected DCAP')
        spec_dcap = self.attestation_specification.intelDcap
        # https://github.com/intel/SGXDataCenterAttestationPrimitives/blob/b6d6145c21e7a452f05838af24b09965ae9b6f10/QuoteGeneration/quote_wrapper/common/inc/sgx_quote_3.h#L177
        quote = dcap.dcapQuote
        certificate = self._dcap_get_cert(quote)
        pck_certs = Pem.parse(certificate)
        # Validate PCK Cert Chain
        trust_roots = [spec_dcap.dcapRootCaDer]
        validation_context = ValidationContext(trust_roots)
        validator = CertificateValidator(pck_certs[0].as_bytes(), intermediate_certs=map(lambda cert: cert.as_bytes(), pck_certs[1:]), validation_context=validation_context)
        validator.validate_usage({"digital_signature"})

        # Verify that the QE REPORT signature
        pubk = asymmetric.dump_public_key(
            asymmetric.load_public_key(certificate))
        vk = VerifyingKey.from_pem(pubk)
        vk.verify(quote[948:1012], quote[564:948], hashfunc=sha256)

        flags = quote[96]
        if not (flags & Verification.IASAttributeFlags.INIT):
            raise EnclaveQuoteFlagsError
        if not (flags & Verification.IASAttributeFlags.MODE64BIT):
            raise EnclaveQuoteFlagsError
        if flags & Verification.IASAttributeFlags.DEBUG:
            if spec_dcap.accept_debug:
                print("!!!WARNING!!! DEBUG quote is being accepted, quote is NOT to be trusted !!!WARNING!!!")
            else:
                raise EnclaveQuoteFlagsError

        # Asserting measurement
        measurement = bytes(quote[112:144])
        if not (spec_dcap.mrenclave == measurement):
            raise MeasurementMismatchError(
                f"Wrong measurement, expected measurement: {spec_dcap.mrenclave.hex()}, actual measurement: {measurement.hex()}."
            )

        tcb_info_dic: TcbInfoContainer = json.loads(dcap.tcbInfo)
        tcb_info_body = bytes(json.dumps(tcb_info_dic["tcbInfo"], separators=(',', ':')), "utf-8")
        tcb_info_sign = bytearray.fromhex(tcb_info_dic["signature"])

        # Verify TCB Sign Cert

        validator_tcb = CertificateValidator(dcap.tcbSignCert, validation_context=validation_context)
        validator_tcb.validate_usage({"digital_signature"})

        # Verify signature over the tcb info body
        tcb_pubk = asymmetric.dump_public_key(asymmetric.load_public_key(dcap.tcbSignCert))
        tcb_vk = VerifyingKey.from_pem(tcb_pubk)
        tcb_vk.verify(tcb_info_sign, tcb_info_body, hashfunc=sha256)

        # Checking the tcbStatus is supported
        cpusvn = [x for x in quote[48:64]]
        pcesvn = struct.unpack("<H", quote[10:12])[0]
        tcb_level = self._dcap_find_tcb_level(tcb_info_dic["tcbInfo"]["tcbLevels"], cpusvn, pcesvn)
        if tcb_level is None:
            raise Exception("TCB level not supported")
        self._dcap_check_status(spec_dcap, tcb_level["tcbStatus"])

        qe_id_dic = json.loads(dcap.qeIdentity)
        qe_id_body = bytes(json.dumps(qe_id_dic["enclaveIdentity"], separators=(',', ':')), "utf-8")
        qe_id_sign = bytearray.fromhex(qe_id_dic["signature"])

        # Verify the Qe Sign Cert
        validator_qe = CertificateValidator(dcap.qeSignCert, validation_context=validation_context)
        validator_qe.validate_usage({"digital_signature"})

        # Verify Signature over enclave identity
        qe_pubk = asymmetric.dump_public_key(asymmetric.load_public_key(dcap.qeSignCert))
        qe_vk = VerifyingKey.from_pem(qe_pubk)
        qe_vk.verify(qe_id_sign, qe_id_body, hashfunc=sha256)

        # Checking if isvprodid is matching
        isvprodid_qe = int.from_bytes(quote[820:822], "little")
        if qe_id_dic["enclaveIdentity"]["isvprodid"] != isvprodid_qe:
            raise Exception(
                f'QE identity of quote {isvprodid_qe} does not match expected identity {qe_id_dic["enclaveIdentity"]["isvprodid"]}')

        isvsvn = int.from_bytes(quote[822:824], "little")

        # Cheking if tcbStatus of the QUoting enclave is supported
        qe_tcb_level = next(
            (level for level in qe_id_dic["enclaveIdentity"]["tcbLevels"] if level["tcb"]["isvsvn"] <= isvsvn), None)
        if qe_tcb_level is None:
            raise Exception("QE TCB level not supported")

        return quote[368:400]

    def verify(self, fatquote: Fatquote) -> bytes:
        """
        :param certificate:
            A byte string of the certificate

        :param message:
            A byte string of the data the signature is for

        :param signature:
            A byte string of the data the signature is for

        :raises:
            CertificateValidationError
            SignatureValidationError
            MessageParsingError
            ISVEnclaveQuoteStatusError
            EnclaveQuoteFlagsError
            MeasurementMismatchError

        :return:
            the QuoteBody inside the message
        """

        if self.check_known_root_ca:
            if fatquote.HasField("epid"):
                _, _, ias_root_ca = asn1crypto.pem.unarmor(intel_sgx_ias_root_ca)
                if ias_root_ca != fatquote.epid.iasRootCaDer:
                    raise Exception(f'Intel IAS Root CA in fatquote does not match known value')
            elif fatquote.HasField("dcap"):
                _, _, dcap_root_ca = asn1crypto.pem.unarmor(intel_sgx_dcap_root_ca)
                if dcap_root_ca != fatquote.dcap.dcapRootCaDer:
                    raise Exception(f'Intel DCAP Root CA in fatquote does not match known value')
            else:
                raise Exception(f'Unknown fatquote type')
        else:
            print("!!!WARNING!!! MOCKED QUOTE is being accepted, quote is NOT to be trusted !!!WARNING!!!")

        if fatquote.HasField("epid"):
            return self._verify_epid(fatquote.epid)
        elif fatquote.HasField("dcap"):
            return self._verify_dcap(fatquote.dcap)
        else:
            raise Exception(f'Unknown fatquote type')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: compute_s3_sink.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x63ompute_s3_sink.proto\x12\x0f\x63ompute_s3_sink\"\xb9\x01\n\x19S3SinkWorkerConfiguration\x12\x10\n\x08\x65ndpoint\x18\x01 \x01(\t\x12\x0e\n\x06region\x18\x02 \x01(\t\x12\x1d\n\x15\x63redentialsDependency\x18\x03 \x01(\t\x12*\n\x07objects\x18\x04 \x03(\x0b\x32\x19.compute_s3_sink.S3Object\x12/\n\ns3Provider\x18\x05 \x01(\x0e\x32\x1b.compute_s3_sink.S3Provider\"~\n\x08S3Object\x12\x12\n\ndependency\x18\x01 \x01(\t\x12)\n\x03zip\x18\x02 \x01(\x0b\x32\x1a.compute_s3_sink.ZipObjectH\x00\x12)\n\x03raw\x18\x03 \x01(\x0b\x32\x1a.compute_s3_sink.RawObjectH\x00\x42\x08\n\x06\x66ormat\"\x18\n\tRawObject\x12\x0b\n\x03key\x18\x01 \x01(\t\"{\n\tZipObject\x12\x31\n\nsingleFile\x18\x01 \x01(\x0b\x32\x1b.compute_s3_sink.SingleFileH\x00\x12\x33\n\x0b\x66ullContent\x18\x02 \x01(\x0b\x32\x1c.compute_s3_sink.FullContentH\x00\x42\x06\n\x04kind\"\'\n\nSingleFile\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\"\r\n\x0b\x46ullContent\"5\n\rS3Credentials\x12\x11\n\taccessKey\x18\x01 \x01(\t\x12\x11\n\tsecretKey\x18\x02 \x01(\t*\x1e\n\nS3Provider\x12\x07\n\x03\x41WS\x10\x00\x12\x07\n\x03GCS\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'compute_s3_sink_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _S3PROVIDER._serialized_start=620
  _S3PROVIDER._serialized_end=650
  _S3SINKWORKERCONFIGURATION._serialized_start=43
  _S3SINKWORKERCONFIGURATION._serialized_end=228
  _S3OBJECT._serialized_start=230
  _S3OBJECT._serialized_end=356
  _RAWOBJECT._serialized_start=358
  _RAWOBJECT._serialized_end=382
  _ZIPOBJECT._serialized_start=384
  _ZIPOBJECT._serialized_end=507
  _SINGLEFILE._serialized_start=509
  _SINGLEFILE._serialized_end=548
  _FULLCONTENT._serialized_start=550
  _FULLCONTENT._serialized_end=563
  _S3CREDENTIALS._serialized_start=565
  _S3CREDENTIALS._serialized_end=618
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class SingleFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___SingleFile = SingleFile

@typing_extensions.final
class RawFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___RawFile = RawFile

@typing_extensions.final
class ZipFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SINGLEFILE_FIELD_NUMBER: builtins.int
    @property
    def singleFile(self) -> global___SingleFile: ...
    def __init__(
        self,
        *,
        singleFile: global___SingleFile | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["selection", b"selection", "singleFile", b"singleFile"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["selection", b"selection", "singleFile", b"singleFile"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["selection", b"selection"]) -> typing_extensions.Literal["singleFile"] | None: ...

global___ZipFile = ZipFile

@typing_extensions.final
class SinkInput(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEPENDENCY_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    RAW_FIELD_NUMBER: builtins.int
    ZIP_FIELD_NUMBER: builtins.int
    dependency: builtins.str
    name: builtins.str
    @property
    def raw(self) -> global___RawFile: ...
    @property
    def zip(self) -> global___ZipFile: ...
    def __init__(
        self,
        *,
        dependency: builtins.str = ...,
        name: builtins.str = ...,
        raw: global___RawFile | None = ...,
        zip: global___ZipFile | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["file", b"file", "raw", b"raw", "zip", b"zip"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dependency", b"dependency", "file", b"file", "name", b"name", "raw", b"raw", "zip", b"zip"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["file", b"file"]) -> typing_extensions.Literal["raw", "zip"] | None: ...

global___SinkInput = SinkInput

@typing_extensions.final
class MetaSinkWorkerConfiguration(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INPUT_FIELD_NUMBER: builtins.int
    ACCESSTOKENDEPENDENCY_FIELD_NUMBER: builtins.int
    ADACCOUNTID_FIELD_NUMBER: builtins.int
    AUDIENCENAME_FIELD_NUMBER: builtins.int
    APIVERSION_FIELD_NUMBER: builtins.int
    @property
    def input(self) -> global___SinkInput: ...
    accessTokenDependency: builtins.str
    """https://graph.facebook.com/v17.0/act_<AD_ACCOUNT_ID>/customaudiences
     Content-Type multipart/form-data
    """
    adAccountId: builtins.str
    audienceName: builtins.str
    apiVersion: builtins.str
    def __init__(
        self,
        *,
        input: global___SinkInput | None = ...,
        accessTokenDependency: builtins.str = ...,
        adAccountId: builtins.str = ...,
        audienceName: builtins.str = ...,
        apiVersion: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_apiVersion", b"_apiVersion", "apiVersion", b"apiVersion", "input", b"input"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_apiVersion", b"_apiVersion", "accessTokenDependency", b"accessTokenDependency", "adAccountId", b"adAccountId", "apiVersion", b"apiVersion", "audienceName", b"audienceName", "input", b"input"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_apiVersion", b"_apiVersion"]) -> typing_extensions.Literal["apiVersion"] | None: ...

global___MetaSinkWorkerConfiguration = MetaSinkWorkerConfiguration



from google.protobuf.json_format import MessageToDict
from ..proto import serialize_length_delimited, ComputeNodeFormat, parse_length_delimited
from ..node import Node
from .proto import (
    GoogleAdManagerWorkerConfiguration,
    SinkInput,
)
from typing import Literal


__docformat__ = "restructuredtext"
__pdoc__ = {
    "proto": False,
}

GoogleAdManagerIdentifierKind = Literal[
                "cookie_encrypted",
                "cookie_idfa",
                "ppid",
                "cookie_rida",
                "cookie_tvos",
                "cookie_adid"]


class GoogleAdManager(Node):
    """
    Compute node that creates a Google Ad Manager segment list.
    """

    def __init__(
            self,
            name: str,
            input: SinkInput,
            credentialsDependency: str,
            identifier_kind: GoogleAdManagerIdentifierKind,
            list_id: str,
            input_has_headers: bool,
            bucket: str,
            object_name: str,
    ) -> None:
        config = GoogleAdManagerWorkerConfiguration(
            input=input,
            credentialsDependency=credentialsDependency,
            identifierKind=identifier_kind,
            listId=list_id,
            inputHasHeaders=input_has_headers,
            bucket=bucket,
            objectName=object_name,
        )
        config_serialized = serialize_length_delimited(config)
        dependencies = [input.dependency, credentialsDependency]
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.google-ad-manager-worker",
            dependencies=dependencies,
            output_format=ComputeNodeFormat.RAW
        )


class GoogleAdManagerWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = GoogleAdManagerWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


__all__ = [
    "GoogleAdManager",
]




from typing import List, Literal
from google.protobuf.json_format import MessageToDict
from .proto.compute_s3_sink_pb2 import S3SinkWorkerConfiguration, S3Object, S3Provider
from ..proto.length_delimited import serialize_length_delimited, parse_length_delimited
from ..proto import ComputeNodeFormat
from ..node import Node


class S3SinkCompute(Node):
    """
    Compute node which allows the upload of files to an S3 api compatible endpoint
    """

    def __init__(
            self,
            name: str,
            endpoint: str,
            credentials_dependency: str,
            s3_provider: Literal["AWS", "GCS"],
            objects: List[S3Object],
            region: str = "",
    ) -> None:
        """
        Create a container compute node.

        **Parameters**:
        - `name`: The name of the node. This serves as an identifier for the
            node and needs to be specified when you interact with the node
            (e.g. run its computation or retrieve its results).
        - `endpoint`: The S3 endpoint including the bucket name.
        - `region`: The S3 region used by the endpoint.
        - `credentials_dependency`: the id of the dependency that holds the credentials
            to access the S3 bucket
            This node should provide a single JSON with format:
            - For AWS:
            ```
            {
                "accessKey": "xxxx",
                "secretKey": "yyyy"
            }
            - For GCS, see [service account credentials](https://cloud.google.com/iam/docs/service-account-creds)
        - `objects`: The list of objects to upload to the S3 bucket
        """
        configuration = S3SinkWorkerConfiguration()
        configuration.endpoint = endpoint
        configuration.region = region
        configuration.credentialsDependency = credentials_dependency
        configuration.objects.extend(objects)
        if s3_provider == "AWS":
            configuration.s3Provider = S3Provider.AWS
        elif s3_provider == "GCS":
            configuration.s3Provider = S3Provider.GCS
        else:
            raise Exception(f"Invalid s3_provider: got {s3_provider}")
        
        config = serialize_length_delimited(configuration)
        dependencies = [credentials_dependency] + list(map(lambda a: a.dependency, objects))

        super().__init__(
            name,
            config=config,
            enclave_type="decentriq.s3-sink-worker",
            dependencies=dependencies,
            output_format=ComputeNodeFormat.RAW
        )


class S3SinkWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = S3SinkWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: permutive.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fpermutive.proto\x12\tpermutive\"\x0c\n\nImportRole\"1\n\nExportRole\x12#\n\x05input\x18\x01 \x01(\x0b\x32\x14.permutive.SinkInput\"\x1a\n\nSingleFile\x12\x0c\n\x04name\x18\x01 \x01(\t\"\t\n\x07RawFile\"C\n\x07ZipFile\x12+\n\nsingleFile\x18\x01 \x01(\x0b\x32\x15.permutive.SingleFileH\x00\x42\x0b\n\tselection\"{\n\tSinkInput\x12\x12\n\ndependency\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12!\n\x03raw\x18\x03 \x01(\x0b\x32\x12.permutive.RawFileH\x00\x12!\n\x03zip\x18\x04 \x01(\x0b\x32\x12.permutive.ZipFileH\x00\x42\x06\n\x04\x66ile\"\xf4\x01\n\x1cPermutiveWorkerConfiguration\x12\x1d\n\x15\x63redentialsDependency\x18\x01 \x01(\t\x12+\n\nimportRole\x18\x02 \x01(\x0b\x32\x15.permutive.ImportRoleH\x00\x12+\n\nexportRole\x18\x03 \x01(\x0b\x32\x15.permutive.ExportRoleH\x00\x12\x10\n\x08importId\x18\x04 \x01(\t\x12\x13\n\x0bsegmentName\x18\x05 \x01(\t\x12\x13\n\x0bsegmentCode\x18\x06 \x01(\t\x12\x17\n\x0finputHasHeaders\x18\x07 \x01(\x08\x42\x06\n\x04roleb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'permutive_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _IMPORTROLE._serialized_start=30
  _IMPORTROLE._serialized_end=42
  _EXPORTROLE._serialized_start=44
  _EXPORTROLE._serialized_end=93
  _SINGLEFILE._serialized_start=95
  _SINGLEFILE._serialized_end=121
  _RAWFILE._serialized_start=123
  _RAWFILE._serialized_end=132
  _ZIPFILE._serialized_start=134
  _ZIPFILE._serialized_end=201
  _SINKINPUT._serialized_start=203
  _SINKINPUT._serialized_end=326
  _PERMUTIVEWORKERCONFIGURATION._serialized_start=329
  _PERMUTIVEWORKERCONFIGURATION._serialized_end=573
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: metering.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import attestation_pb2 as attestation__pb2
from . import data_room_pb2 as data__room__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0emetering.proto\x12\x08metering\x1a\x11\x61ttestation.proto\x1a\x0f\x64\x61ta_room.proto\"[\n\x0b\x45nclaveInfo\x12\x1e\n\x16\x61ttestationSpecHashHex\x18\x01 \x01(\t\x12\x1a\n\rtaskQueueName\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_taskQueueName\"\xb9\x05\n\x0fMeteringRequest\x12/\n\tcreateDcr\x18\x01 \x01(\x0b\x32\x1a.metering.CreateDcrRequestH\x00\x12;\n\x0f\x63reateDcrCommit\x18\x02 \x01(\x0b\x32 .metering.CreateDcrCommitRequestH\x00\x12+\n\x07stopDcr\x18\x03 \x01(\x0b\x32\x18.metering.StopDcrRequestH\x00\x12\x39\n\x0epublishDataset\x18\x04 \x01(\x0b\x32\x1f.metering.PublishDatasetRequestH\x00\x12=\n\x10unpublishDataset\x18\x05 \x01(\x0b\x32!.metering.UnpublishDatasetRequestH\x00\x12\x39\n\x0eworkerMetadata\x18\x06 \x01(\x0b\x32\x1f.metering.WorkerMetadataRequestH\x00\x12O\n\x19submitWorkerExecutionTime\x18\x07 \x01(\x0b\x32*.metering.SubmitWorkerExecutionTimeRequestH\x00\x12\x39\n\x0e\x64\x63rInteraction\x18\x08 \x01(\x0b\x32\x1f.metering.DcrInteractionRequestH\x00\x12\x37\n\rcreateDataset\x18\t \x01(\x0b\x32\x1e.metering.CreateDatasetRequestH\x00\x12K\n\x17getOrCreateDatasetScope\x18\n \x01(\x0b\x32(.metering.GetOrCreateDatasetScopeRequestH\x00\x12\x39\n\x0emergeDcrCommit\x18\x0b \x01(\x0b\x32\x1f.metering.MergeDcrCommitRequestH\x00\x42\t\n\x07request\"\xcd\x05\n\x17MeteringSuccessResponse\x12\x30\n\tcreateDcr\x18\x01 \x01(\x0b\x32\x1b.metering.CreateDcrResponseH\x00\x12<\n\x0f\x63reateDcrCommit\x18\x02 \x01(\x0b\x32!.metering.CreateDcrCommitResponseH\x00\x12,\n\x07stopDcr\x18\x03 \x01(\x0b\x32\x19.metering.StopDcrResponseH\x00\x12:\n\x0epublishDataset\x18\x04 \x01(\x0b\x32 .metering.PublishDatasetResponseH\x00\x12>\n\x10unpublishDataset\x18\x05 \x01(\x0b\x32\".metering.UnpublishDatasetResponseH\x00\x12:\n\x0eworkerMetadata\x18\x06 \x01(\x0b\x32 .metering.WorkerMetadataResponseH\x00\x12P\n\x19submitWorkerExecutionTime\x18\x07 \x01(\x0b\x32+.metering.SubmitWorkerExecutionTimeResponseH\x00\x12:\n\x0e\x64\x63rInteraction\x18\x08 \x01(\x0b\x32 .metering.DcrInteractionResponseH\x00\x12\x38\n\rcreateDataset\x18\t \x01(\x0b\x32\x1f.metering.CreateDatasetResponseH\x00\x12L\n\x17getOrCreateDatasetScope\x18\n \x01(\x0b\x32).metering.GetOrCreateDatasetScopeResponseH\x00\x12:\n\x0emergeDcrCommit\x18\x0b \x01(\x0b\x32 .metering.MergeDcrCommitResponseH\x00\x42\n\n\x08response\"g\n\x10MeteringResponse\x12\x34\n\x07success\x18\x01 \x01(\x0b\x32!.metering.MeteringSuccessResponseH\x00\x12\x11\n\x07\x66\x61ilure\x18\x02 \x01(\tH\x00\x42\n\n\x08response\"\x98\x01\n\x0b\x44\x63rMetadata\x12+\n\x07purpose\x18\x01 \x01(\x0e\x32\x1a.metering.CreateDcrPurpose\x12\x1c\n\x14showOrganizationLogo\x18\x02 \x01(\x08\x12\x17\n\x0frequirePassword\x18\x03 \x01(\x08\x12%\n\x04kind\x18\x04 \x01(\x0e\x32\x17.metering.CreateDcrKind\"\xb9\x01\n\x10\x43reateDcrRequest\x12\r\n\x05idHex\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x03 \x01(\t\x12\x19\n\x11participantEmails\x18\x04 \x03(\t\x12\x15\n\x08metadata\x18\x05 \x01(\x0cH\x00\x88\x01\x01\x12*\n\x0b\x65nclaveInfo\x18\x06 \x03(\x0b\x32\x15.metering.EnclaveInfoB\x0b\n\t_metadata\"~\n!ExecuteComputationNodeInteraction\x12\x0e\n\x06nodeId\x18\x01 \x01(\t\x12\x38\n\x0crateLimiting\x18\x02 \x01(\x0b\x32\x1d.data_room.RateLimitingConfigH\x00\x88\x01\x01\x42\x0f\n\r_rateLimiting\"[\n\x1d\x45xecuteComputationInteraction\x12:\n\x05nodes\x18\x01 \x03(\x0b\x32+.metering.ExecuteComputationNodeInteraction\"v\n\x19PublishDatasetInteraction\x12\x0e\n\x06nodeId\x18\x01 \x01(\t\x12\x38\n\x0crateLimiting\x18\x02 \x01(\x0b\x32\x1d.data_room.RateLimitingConfigH\x00\x88\x01\x01\x42\x0f\n\r_rateLimiting\"\xa2\x01\n\x12\x44\x63rInteractionKind\x12\x45\n\x12\x65xecuteComputation\x18\x01 \x01(\x0b\x32\'.metering.ExecuteComputationInteractionH\x00\x12=\n\x0epublishDataset\x18\x02 \x01(\x0b\x32#.metering.PublishDatasetInteractionH\x00\x42\x06\n\x04kind\"\x90\x01\n\x15\x44\x63rInteractionRequest\x12\x14\n\x0c\x64\x61taRoomHash\x18\x01 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x02 \x01(\t\x12\x0f\n\x07scopeId\x18\x03 \x01(\t\x12\x31\n\x0binteraction\x18\x04 \x01(\x0b\x32\x1c.metering.DcrInteractionKind\"y\n\x15PublishDatasetRequest\x12\x15\n\rcomputeNodeId\x18\x01 \x01(\t\x12\x14\n\x0cmanifestHash\x18\x02 \x01(\t\x12\x14\n\x0c\x64\x61taRoomHash\x18\x03 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x04 \x01(\t\"\x99\x02\n\x14\x43reateDatasetRequest\x12\x14\n\x0cmanifestHash\x18\x01 \x01(\t\x12\x15\n\x08manifest\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x0f\n\x07scopeId\x18\x03 \x01(\t\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\x18\n\x0b\x64\x65scription\x18\x05 \x01(\tH\x01\x88\x01\x01\x12\x16\n\tsizeBytes\x18\x06 \x01(\x04H\x02\x88\x01\x01\x12\x17\n\nstatistics\x18\x07 \x01(\tH\x03\x88\x01\x01\x12\x1c\n\x0f\x64\x61tasetImportId\x18\x08 \x01(\tH\x04\x88\x01\x01\x42\x0b\n\t_manifestB\x0e\n\x0c_descriptionB\x0c\n\n_sizeBytesB\r\n\x0b_statisticsB\x12\n\x10_datasetImportId\"L\n\x1eGetOrCreateDatasetScopeRequest\x12\x19\n\x0cmanifestHash\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x0f\n\r_manifestHash\"e\n\x17UnpublishDatasetRequest\x12\x15\n\rcomputeNodeId\x18\x01 \x01(\t\x12\x14\n\x0c\x64\x61taRoomHash\x18\x02 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x03 \x01(\t\"E\n\x0eStopDcrRequest\x12\x14\n\x0c\x64\x61taRoomHash\x18\x01 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x02 \x01(\t\"\x81\x01\n\x16\x43reateDcrCommitRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08\x64\x63rIdHex\x18\x02 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x03 \x01(\t\x12*\n\x0b\x65nclaveInfo\x18\x06 \x03(\x0b\x32\x15.metering.EnclaveInfo\"\x80\x01\n\x15MergeDcrCommitRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08\x64\x63rIdHex\x18\x02 \x01(\t\x12\x1d\n\x15\x64riverAttestationHash\x18\x03 \x01(\t\x12*\n\x0b\x65nclaveInfo\x18\x06 \x03(\x0b\x32\x15.metering.EnclaveInfo\"\x13\n\x11\x43reateDcrResponse\"-\n\x1fGetOrCreateDatasetScopeResponse\x12\n\n\x02id\x18\x01 \x01(\t\"\x18\n\x16\x44\x63rInteractionResponse\"\x19\n\x17\x43reateDcrCommitResponse\"\x18\n\x16MergeDcrCommitResponse\"\x11\n\x0fStopDcrResponse\"\x18\n\x16PublishDatasetResponse\"#\n\x15\x43reateDatasetResponse\x12\n\n\x02id\x18\x01 \x01(\t\"\x1a\n\x18UnpublishDatasetResponse\"h\n\x15WorkerMetadataRequest\x12>\n\x0f\x61ttestationSpec\x18\x01 \x01(\x0b\x32%.attestation.AttestationSpecification\x12\x0f\n\x07scopeId\x18\x02 \x01(\t\"5\n\x16WorkerMetadataResponse\x12\x1b\n\x13maxExecutionSeconds\x18\x01 \x01(\x04\"\x91\x01\n SubmitWorkerExecutionTimeRequest\x12\x1c\n\x14\x65xecutionTimeSeconds\x18\x01 \x01(\r\x12>\n\x0f\x61ttestationSpec\x18\x02 \x01(\x0b\x32%.attestation.AttestationSpecification\x12\x0f\n\x07scopeId\x18\x03 \x01(\t\"#\n!SubmitWorkerExecutionTimeResponse*`\n\x10\x43reateDcrPurpose\x12\x0c\n\x08STANDARD\x10\x00\x12\x0e\n\nVALIDATION\x10\x01\x12\x0f\n\x0b\x44\x41TA_IMPORT\x10\x02\x12\x0f\n\x0b\x44\x41TA_EXPORT\x10\x03\x12\x0c\n\x08\x44\x41TA_LAB\x10\x04*L\n\rCreateDcrKind\x12\n\n\x06\x45XPERT\x10\x00\x12\x0f\n\x0b\x44\x41TASCIENCE\x10\x01\x12\t\n\x05MEDIA\x10\x02\x12\x13\n\x0fLOOKALIKE_MEDIA\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'metering_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CREATEDCRPURPOSE._serialized_start=4203
  _CREATEDCRPURPOSE._serialized_end=4299
  _CREATEDCRKIND._serialized_start=4301
  _CREATEDCRKIND._serialized_end=4377
  _ENCLAVEINFO._serialized_start=64
  _ENCLAVEINFO._serialized_end=155
  _METERINGREQUEST._serialized_start=158
  _METERINGREQUEST._serialized_end=855
  _METERINGSUCCESSRESPONSE._serialized_start=858
  _METERINGSUCCESSRESPONSE._serialized_end=1575
  _METERINGRESPONSE._serialized_start=1577
  _METERINGRESPONSE._serialized_end=1680
  _DCRMETADATA._serialized_start=1683
  _DCRMETADATA._serialized_end=1835
  _CREATEDCRREQUEST._serialized_start=1838
  _CREATEDCRREQUEST._serialized_end=2023
  _EXECUTECOMPUTATIONNODEINTERACTION._serialized_start=2025
  _EXECUTECOMPUTATIONNODEINTERACTION._serialized_end=2151
  _EXECUTECOMPUTATIONINTERACTION._serialized_start=2153
  _EXECUTECOMPUTATIONINTERACTION._serialized_end=2244
  _PUBLISHDATASETINTERACTION._serialized_start=2246
  _PUBLISHDATASETINTERACTION._serialized_end=2364
  _DCRINTERACTIONKIND._serialized_start=2367
  _DCRINTERACTIONKIND._serialized_end=2529
  _DCRINTERACTIONREQUEST._serialized_start=2532
  _DCRINTERACTIONREQUEST._serialized_end=2676
  _PUBLISHDATASETREQUEST._serialized_start=2678
  _PUBLISHDATASETREQUEST._serialized_end=2799
  _CREATEDATASETREQUEST._serialized_start=2802
  _CREATEDATASETREQUEST._serialized_end=3083
  _GETORCREATEDATASETSCOPEREQUEST._serialized_start=3085
  _GETORCREATEDATASETSCOPEREQUEST._serialized_end=3161
  _UNPUBLISHDATASETREQUEST._serialized_start=3163
  _UNPUBLISHDATASETREQUEST._serialized_end=3264
  _STOPDCRREQUEST._serialized_start=3266
  _STOPDCRREQUEST._serialized_end=3335
  _CREATEDCRCOMMITREQUEST._serialized_start=3338
  _CREATEDCRCOMMITREQUEST._serialized_end=3467
  _MERGEDCRCOMMITREQUEST._serialized_start=3470
  _MERGEDCRCOMMITREQUEST._serialized_end=3598
  _CREATEDCRRESPONSE._serialized_start=3600
  _CREATEDCRRESPONSE._serialized_end=3619
  _GETORCREATEDATASETSCOPERESPONSE._serialized_start=3621
  _GETORCREATEDATASETSCOPERESPONSE._serialized_end=3666
  _DCRINTERACTIONRESPONSE._serialized_start=3668
  _DCRINTERACTIONRESPONSE._serialized_end=3692
  _CREATEDCRCOMMITRESPONSE._serialized_start=3694
  _CREATEDCRCOMMITRESPONSE._serialized_end=3719
  _MERGEDCRCOMMITRESPONSE._serialized_start=3721
  _MERGEDCRCOMMITRESPONSE._serialized_end=3745
  _STOPDCRRESPONSE._serialized_start=3747
  _STOPDCRRESPONSE._serialized_end=3764
  _PUBLISHDATASETRESPONSE._serialized_start=3766
  _PUBLISHDATASETRESPONSE._serialized_end=3790
  _CREATEDATASETRESPONSE._serialized_start=3792
  _CREATEDATASETRESPONSE._serialized_end=3827
  _UNPUBLISHDATASETRESPONSE._serialized_start=3829
  _UNPUBLISHDATASETRESPONSE._serialized_end=3855
  _WORKERMETADATAREQUEST._serialized_start=3857
  _WORKERMETADATAREQUEST._serialized_end=3961
  _WORKERMETADATARESPONSE._serialized_start=3963
  _WORKERMETADATARESPONSE._serialized_end=4016
  _SUBMITWORKEREXECUTIONTIMEREQUEST._serialized_start=4019
  _SUBMITWORKEREXECUTIONTIMEREQUEST._serialized_end=4164
  _SUBMITWORKEREXECUTIONTIMERESPONSE._serialized_start=4166
  _SUBMITWORKEREXECUTIONTIMERESPONSE._serialized_end=4201
# @@protoc_insertion_point(module_scope)



from typing import List, Optional
from google.protobuf.json_format import MessageToDict
from .proto.compute_container_pb2 import ContainerWorkerConfiguration, MountPoint, ProxyConfiguration
from ..proto.length_delimited import serialize_length_delimited, parse_length_delimited
from ..proto import ComputeNodeFormat
from ..node import Node


class StaticContainerCompute(Node):
    """
    Compute node which allows the execution of programs inside a fixed
    container image.
    """

    def __init__(
            self,
            name: str,
            command: List[str],
            mount_points: List[MountPoint],
            output_path: str,
            enclave_type: str,
            include_container_logs_on_error: bool = False,
            include_container_logs_on_success: bool = False,
            proxy_configuration: Optional[ProxyConfiguration] = None,
    ) -> None:
        """
        Create a container compute node.

        **Parameters**:
        - `name`: The name of the node. This serves as an identifier for the
            node and needs to be specified when you interact with the node
            (e.g. run its computation or retrieve its results).
        - `command`: The command to execute within the container.
        - `mount_points`: A list of mount points that tell the enclave
            at which file system paths which input should be mounted
            (e.g. the contents of a data node or the output of another
            compute node).
        - `output_path`: A path to a directory that will contain all the
            output written by the command executed in this container.
            Files within this directory will be zipped and downloadable.
        - `enclave_type`: The particular enclave to use for this container.
            This setting controls the environment in which the given `command`
            will be run, i.e. what programs and libraries are available.
            This identifier corresponds to the worker name without the version suffix,
            i.e. `"decentriq.python-ml-worker-32-64"`.
        - `include_container_logs_on_error`: Whether to report the internal
            container logs to the outside in case of an error. These logs
            could contain sensitive data and therefore this setting should
            only be used for debugging.
        - `include_container_logs_on_success`: Whether to report the internal
            container logs as part of the result zip file.
            Note that these logs could contain sensitive data and therefore this
            setting should only be used for debugging.
        """
        configuration = ContainerWorkerConfiguration()
        configuration.static.command.extend(command)
        configuration.static.mountPoints.extend(mount_points)
        configuration.static.outputPath = output_path
        configuration.static.includeContainerLogsOnError = include_container_logs_on_error
        configuration.static.includeContainerLogsOnSuccess = include_container_logs_on_success
        if proxy_configuration is not None:
            configuration.static.proxyConfiguration.MergeFrom(proxy_configuration)
        config = serialize_length_delimited(configuration)
        dependencies = list(map(lambda a: a.dependency, mount_points))

        super().__init__(
            name,
            config=config,
            enclave_type=enclave_type,
            dependencies=dependencies,
            output_format=ComputeNodeFormat.ZIP
        )


class ContainerWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = ContainerWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)



from __future__ import annotations
import typing
from typing import Dict, Literal, Optional, Union, Tuple, List
from .client import Client, KeychainInstance
import cbor2
import chily
from base64 import b64encode

KeychainEntryKind = Literal["dataset_key", "other_secret"]

class KeychainEntry:
    kind: KeychainEntryKind
    key: str
    value: bytes

    def __init__(self, kind: KeychainEntryKind, key: str, value: bytes):
        self.kind = kind
        self.key = key
        self.value = value

class KeychainDecryptException(Exception):
    def __init__(self) -> None:
        super().__init__()

class Keychain:

    _client: Client
    _store: Dict[str, bytes]
    _secret_wrapper: chily.SecretWrapper
    _keychain_instance: KeychainInstance

    def __init__(
            self,
            client: Client,
            secret_wrapper: chily.SecretWrapper,
            keychain_instance: KeychainInstance,
            store: Dict[str, bytes],
    ):
        self._client = client
        self._store = store
        self._secret_wrapper = secret_wrapper
        self._keychain_instance = keychain_instance

    @staticmethod
    def _serialize_store(store: Dict[str, bytes]) -> bytes:
        return cbor2.dumps(store, value_sharing=False)

    @staticmethod
    def _deserialize_store(plaintext: bytes) -> Dict[str, bytes]:
        return cbor2.loads(plaintext)

    def _namespaced_key(self, kind: KeychainEntryKind, key: str) -> str:
        return f"{kind}/{key}"

    @staticmethod
    def _parse_namespaced_key(ns_key: str) -> Tuple[KeychainEntryKind, str]:
        splitted = ns_key.split("/")
        if len(splitted) < 2:
            raise Exception("Invalid namespaced key: no `/` found")
        if splitted[0] not in typing.get_args(KeychainEntryKind):
            raise Exception(f"Invalid namespaced key kind: found {splitted[0]}")
        remainder = "/".join(splitted[1:])
        return [splitted[0], remainder]

    @staticmethod
    def _encrypt_store_static(
            user_email: str,
            secret_wrapper: chily.SecretWrapper,
            store: Dict[str, bytes]
    ) -> bytes:
        return secret_wrapper.wrap_secret(user_email, Keychain._serialize_store(store))

    def _encrypt_store(
            self,
    ):
        self._keychain_instance["encrypted"] = Keychain._encrypt_store_static(
            self._client.user_email,
            self._secret_wrapper,
            self._store
        )

    @staticmethod
    def _decrypt_store_static(
            user_email: str,
            secret_wrapper: chily.SecretWrapper,
            encrypted: bytes
    ) -> Dict[str, bytes]:
        try:
            decrypted = secret_wrapper.unwrap_secret(user_email, encrypted)
        except:
            raise KeychainDecryptException()
        return Keychain._deserialize_store(decrypted)

    def _decrypt_store(self):
        self._store = Keychain._decrypt_store_static(
            self._client.user_email,
            self._secret_wrapper,
            self._keychain_instance["encrypted"]
        )

    def _set_keychain_instance(self, keychain_instance: KeychainInstance):
        self._keychain_instance = keychain_instance
        self._decrypt_store()

    def _set_store(self, store: Dict[str, bytes]):
        self._store = store
        self._encrypt_store()

    @staticmethod
    def _init_keychain_instance(client: Client) -> KeychainInstance:
        keychain_instance = client.get_keychain_instance()
        if not keychain_instance:
            raise Exception("Cannot init Keychain, no instance exists in the server, create a new instance first")
        return keychain_instance

    @staticmethod
    def _has_keychain_instance(client: Client) -> bool:
        keychain_instance = client.get_keychain_instance()
        return bool(keychain_instance)

    @staticmethod
    def _create_new_keychain_with_secret_wrapper(
            client: Client,
            secret_wrapper: chily.SecretWrapper,
            check_for_existing_keychain: bool = True
    ) -> Optional[Keychain]:
        if check_for_existing_keychain and Keychain._has_keychain_instance(client):
            return None
        store: Dict[str, bytes] = {}
        encrypted = secret_wrapper.wrap_secret(client.user_email, Keychain._serialize_store(store))
        keychain_instance = client.create_keychain_instance(secret_wrapper.salt, encrypted)
        return Keychain(client, secret_wrapper, keychain_instance, store)

    @staticmethod
    def create_new_keychain(
            client: Client,
            password: bytes,
            check_for_existing_keychain: bool = True,
    ) -> Optional[Keychain]:
        """
        Create a new keychain that is encrypted using the given password.

        If the user already has a keychain setup (for example by already having
        logged into the Decentriq UI), this method will return None.

        See the method `get_or_create_unlocked_keychain` for a convenience
        method that will not throw an error if the keychain already exists.
        """
        secret_wrapper = chily.SecretWrapper.init(password)
        return Keychain._create_new_keychain_with_secret_wrapper(
            client, secret_wrapper, check_for_existing_keychain
        )

    @staticmethod
    def get_or_create_unlocked_keychain(
            client: Client,
            password: bytes,
    ) -> Keychain:
        """
        Get and unlock the user's keychain using the provided password.

        If the user did not already create a keychain, a new keychain will
        be created automatically.
        If a keychain exists but the provided password does not match,
        an exception will be thrown.

        Note that the password must be given as a `bytes` object.
        """
        keychain_instance = client.get_keychain_instance()
        if not keychain_instance:
            keychain = Keychain.create_new_keychain(
                client,
                password,
                check_for_existing_keychain=False,
            )
            if keychain:
                keychain._decrypt_store()
            else:
                raise Exception(
                    "Received an undefined keychain even though it should exist"
                )
            return keychain
        else:
            return Keychain._decrypt_store_with_password(
                client, keychain_instance, password
            )

    @staticmethod
    def create_new_keychain_with_master_key(
            client: Client,
            master_key: bytes,
            salt: str,
    ) -> Optional[Keychain]:
        """
        Create a new keychain with the given master key.

        If the user already has a keychain setup (for example by already having
        logged into the Decentriq UI), this method will return None.
        """
        secret_wrapper = chily.SecretWrapper.with_master_key(master_key, salt)
        return Keychain._create_new_keychain_with_secret_wrapper(client, secret_wrapper)

    @staticmethod
    def init_with_master_key(client: Client, master_key: bytes) -> Keychain:
        """
        Decrypt an existing keychain with the given master key.

        If no keychain has been created already or if the key
        does not match the keychain, an error will be thrown.

        See the method `get_or_create_unlocked_keychain` for a convenience
        method that will create the keychain if it does not exist already.
        """
        keychain_instance = Keychain._init_keychain_instance(client)
        secret_wrapper = chily.SecretWrapper.with_master_key(master_key, keychain_instance["salt"])
        store = Keychain._decrypt_store_static(client.user_email, secret_wrapper, keychain_instance["encrypted"])
        return Keychain(client, secret_wrapper, keychain_instance, store)

    @staticmethod
    def init_with_password(client: Client, password: bytes) -> Keychain:
        """
        Decrypt an existing keychain with the given password.

        If no keychain has been created already or if the password
        does not match the keychain, an error will be thrown.

        See the method `get_or_create_unlocked_keychain` for a convenience
        method that will create the keychain if it does not exist already.
        """
        keychain_instance = Keychain._init_keychain_instance(client)
        return Keychain._decrypt_store_with_password(client, keychain_instance, password)

    @staticmethod
    def _decrypt_store_with_password(
            client: Client,
            keychain_instance: KeychainInstance,
            password: bytes
    ) -> Keychain:
        secret_wrapper = chily.SecretWrapper.with_password(
            password, keychain_instance["salt"]
        )
        store = Keychain._decrypt_store_static(
            client.user_email, secret_wrapper, keychain_instance["encrypted"]
        )
        return Keychain(client, secret_wrapper, keychain_instance, store)

    def _download(self):
        keychain_instance = self._client.get_keychain_instance()
        if not keychain_instance:
            raise Exception("Keychain instance has not been created yet")

        if self._secret_wrapper.salt != keychain_instance["salt"]:
            raise Exception("Keychain salt has changed upstream. Reinitialize Keychain with password first")

        if keychain_instance["casIndex"] != self._keychain_instance["casIndex"]:
            self._set_keychain_instance(keychain_instance)
        return

    def get(self, kind: KeychainEntryKind, key: str) -> Optional[KeychainEntry]:
        self._download()
        ns_key = self._namespaced_key(kind, key)
        value = self._store.get(ns_key)
        if value:
            return KeychainEntry(kind, key, value)
        else:
            return None

    def items(self) -> List[KeychainEntry]:
        self._download()
        items = []
        for ns_key, value in self._store.items():
            kind, key = Keychain._parse_namespaced_key(ns_key)
            items.append(KeychainEntry(kind, key, value))
        return items

    def _insert_local(self, entry: KeychainEntry):
        ns_key = self._namespaced_key(entry.kind, entry.key)
        if self._store.get(ns_key):
            raise Exception("Cannot insert new entry: an entry already exists for this (kind, key) pair")
        if entry.kind in ["dataset_key", "other_secret"]:
            self._store[ns_key] = entry.value
        else:
            raise Exception(f"Invalid entry kind: {entry.kind}")
        self._encrypt_store()
        return

    def _compare_and_swap(self) -> bool:
        return self._client.compare_and_swap_keychain(
            self._keychain_instance["casIndex"],
            self._keychain_instance["salt"],
            self._keychain_instance["encrypted"],
        )

    def insert(self, entry: KeychainEntry):
        while True:
            self._insert_local(entry)
            if self._compare_and_swap():
                return
            self._download()

    def remove(self, kind: KeychainEntryKind, key: str):
        ns_key = self._namespaced_key(kind, key)
        while True:
            if self._store.pop(ns_key):
                self._encrypt_store()
                if self._compare_and_swap():
                    return
                self._download()
            else:
                raise Exception("Cannot delete entry: Entry does not exist")

    @staticmethod
    def reset(client: Client):
        return client.reset_keychain()

    def clear(self):
        while True:
            self._set_store({})
            if self._compare_and_swap():
                return
            self._download()

    def change_password(self, new_password: bytes):
        new_secret_wrapper = chily.SecretWrapper.with_password(new_password, self._secret_wrapper.salt)
        return self.change_master_key(new_secret_wrapper.master_key)

    def change_master_key(self, new_master_key: bytes):
        new_secret_wrapper = chily.SecretWrapper.with_master_key(new_master_key, self._secret_wrapper.salt)
        while True:
            new_keychain = Keychain(self._client, new_secret_wrapper, self._keychain_instance, self._store)
            new_keychain._encrypt_store()
            if new_keychain._compare_and_swap():
                self.__dict__.update(new_keychain.__dict__)
                return
            self._download()

    def get_master_key(self) -> bytes:
        return self._secret_wrapper.master_key



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gcg.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import data_room_pb2 as data__room__pb2
from . import identity_endorsement_pb2 as identity__endorsement__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tgcg.proto\x12\x03gcg\x1a\x0f\x64\x61ta_room.proto\x1a\x1aidentity_endorsement.proto\"\x82\x0e\n\nGcgRequest\x12\x1f\n\x08userAuth\x18\x01 \x01(\x0b\x32\r.gcg.UserAuth\x12;\n\x15\x63reateDataRoomRequest\x18\x02 \x01(\x0b\x32\x1a.gcg.CreateDataRoomRequestH\x00\x12?\n\x17retrieveDataRoomRequest\x18\x03 \x01(\x0b\x32\x1c.gcg.RetrieveDataRoomRequestH\x00\x12g\n+retrieveCurrentDataRoomConfigurationRequest\x18\x04 \x01(\x0b\x32\x30.gcg.RetrieveCurrentDataRoomConfigurationRequestH\x00\x12K\n\x1dretrieveDataRoomStatusRequest\x18\x05 \x01(\x0b\x32\".gcg.RetrieveDataRoomStatusRequestH\x00\x12G\n\x1bupdateDataRoomStatusRequest\x18\x06 \x01(\x0b\x32 .gcg.UpdateDataRoomStatusRequestH\x00\x12?\n\x17retrieveAuditLogRequest\x18\x07 \x01(\x0b\x32\x1c.gcg.RetrieveAuditLogRequestH\x00\x12O\n\x1fpublishDatasetToDataRoomRequest\x18\x08 \x01(\x0b\x32$.gcg.PublishDatasetToDataRoomRequestH\x00\x12Q\n retrievePublishedDatasetsRequest\x18\t \x01(\x0b\x32%.gcg.RetrievePublishedDatasetsRequestH\x00\x12K\n\x1dremovePublishedDatasetRequest\x18\n \x01(\x0b\x32\".gcg.RemovePublishedDatasetRequestH\x00\x12;\n\x15\x65xecuteComputeRequest\x18\x0b \x01(\x0b\x32\x1a.gcg.ExecuteComputeRequestH\x00\x12\x31\n\x10jobStatusRequest\x18\x0c \x01(\x0b\x32\x15.gcg.JobStatusRequestH\x00\x12\x33\n\x11getResultsRequest\x18\r \x01(\x0b\x32\x16.gcg.GetResultsRequestH\x00\x12Q\n createConfigurationCommitRequest\x18\x0e \x01(\x0b\x32%.gcg.CreateConfigurationCommitRequestH\x00\x12U\n\"retrieveConfigurationCommitRequest\x18\x0f \x01(\x0b\x32\'.gcg.RetrieveConfigurationCommitRequestH\x00\x12Q\n executeDevelopmentComputeRequest\x18\x10 \x01(\x0b\x32%.gcg.ExecuteDevelopmentComputeRequestH\x00\x12[\n%generateMergeApprovalSignatureRequest\x18\x11 \x01(\x0b\x32*.gcg.GenerateMergeApprovalSignatureRequestH\x00\x12O\n\x1fmergeConfigurationCommitRequest\x18\x12 \x01(\x0b\x32$.gcg.MergeConfigurationCommitRequestH\x00\x12g\n+retrieveConfigurationCommitApproversRequest\x18\x13 \x01(\x0b\x32\x30.gcg.RetrieveConfigurationCommitApproversRequestH\x00\x12\x41\n\x18\x63\x61sAuxiliaryStateRequest\x18\x14 \x01(\x0b\x32\x1d.gcg.CasAuxiliaryStateRequestH\x00\x12\x43\n\x19readAuxiliaryStateRequest\x18\x15 \x01(\x0b\x32\x1e.gcg.ReadAuxiliaryStateRequestH\x00\x12O\n\x1fretrieveUsedAirlockQuotaRequest\x18\x16 \x01(\x0b\x32$.gcg.RetrieveUsedAirlockQuotaRequestH\x00\x12;\n\x15getResultsSizeRequest\x18\x17 \x01(\x0b\x32\x1a.gcg.GetResultsSizeRequestH\x00\x12\x46\n\x12\x65ndorsementRequest\x18\x65 \x01(\x0b\x32(.identity_endorsement.EndorsementRequestH\x00\x42\r\n\x0bgcg_request\"\x9c\x0e\n\x0bGcgResponse\x12\x11\n\x07\x66\x61ilure\x18\x01 \x01(\tH\x00\x12=\n\x16\x63reateDataRoomResponse\x18\x02 \x01(\x0b\x32\x1b.gcg.CreateDataRoomResponseH\x00\x12\x41\n\x18retrieveDataRoomResponse\x18\x03 \x01(\x0b\x32\x1d.gcg.RetrieveDataRoomResponseH\x00\x12i\n,retrieveCurrentDataRoomConfigurationResponse\x18\x04 \x01(\x0b\x32\x31.gcg.RetrieveCurrentDataRoomConfigurationResponseH\x00\x12M\n\x1eretrieveDataRoomStatusResponse\x18\x05 \x01(\x0b\x32#.gcg.RetrieveDataRoomStatusResponseH\x00\x12I\n\x1cupdateDataRoomStatusResponse\x18\x06 \x01(\x0b\x32!.gcg.UpdateDataRoomStatusResponseH\x00\x12\x41\n\x18retrieveAuditLogResponse\x18\x07 \x01(\x0b\x32\x1d.gcg.RetrieveAuditLogResponseH\x00\x12Q\n publishDatasetToDataRoomResponse\x18\x08 \x01(\x0b\x32%.gcg.PublishDatasetToDataRoomResponseH\x00\x12S\n!retrievePublishedDatasetsResponse\x18\t \x01(\x0b\x32&.gcg.RetrievePublishedDatasetsResponseH\x00\x12M\n\x1eremovePublishedDatasetResponse\x18\n \x01(\x0b\x32#.gcg.RemovePublishedDatasetResponseH\x00\x12=\n\x16\x65xecuteComputeResponse\x18\x0b \x01(\x0b\x32\x1b.gcg.ExecuteComputeResponseH\x00\x12\x33\n\x11jobStatusResponse\x18\x0c \x01(\x0b\x32\x16.gcg.JobStatusResponseH\x00\x12?\n\x17getResultsResponseChunk\x18\r \x01(\x0b\x32\x1c.gcg.GetResultsResponseChunkH\x00\x12\x41\n\x18getResultsResponseFooter\x18\x0e \x01(\x0b\x32\x1d.gcg.GetResultsResponseFooterH\x00\x12S\n!createConfigurationCommitResponse\x18\x0f \x01(\x0b\x32&.gcg.CreateConfigurationCommitResponseH\x00\x12W\n#retrieveConfigurationCommitResponse\x18\x10 \x01(\x0b\x32(.gcg.RetrieveConfigurationCommitResponseH\x00\x12]\n&generateMergeApprovalSignatureResponse\x18\x11 \x01(\x0b\x32+.gcg.GenerateMergeApprovalSignatureResponseH\x00\x12Q\n mergeConfigurationCommitResponse\x18\x12 \x01(\x0b\x32%.gcg.MergeConfigurationCommitResponseH\x00\x12i\n,retrieveConfigurationCommitApproversResponse\x18\x13 \x01(\x0b\x32\x31.gcg.RetrieveConfigurationCommitApproversResponseH\x00\x12\x43\n\x19\x63\x61sAuxiliaryStateResponse\x18\x14 \x01(\x0b\x32\x1e.gcg.CasAuxiliaryStateResponseH\x00\x12\x45\n\x1areadAuxiliaryStateResponse\x18\x15 \x01(\x0b\x32\x1f.gcg.ReadAuxiliaryStateResponseH\x00\x12Q\n retrieveUsedAirlockQuotaResponse\x18\x16 \x01(\x0b\x32%.gcg.RetrieveUsedAirlockQuotaResponseH\x00\x12=\n\x16getResultsSizeResponse\x18\x17 \x01(\x0b\x32\x1b.gcg.GetResultsSizeResponseH\x00\x12H\n\x13\x65ndorsementResponse\x18\x65 \x01(\x0b\x32).identity_endorsement.EndorsementResponseH\x00\x42\x0e\n\x0cgcg_response\"i\n\x08UserAuth\x12\x15\n\x03pki\x18\x01 \x01(\x0b\x32\x08.gcg.Pki\x12\x46\n\x13\x65nclaveEndorsements\x18\x02 \x01(\x0b\x32).identity_endorsement.EnclaveEndorsements\"=\n\x03Pki\x12\x14\n\x0c\x63\x65rtChainPem\x18\x01 \x01(\x0c\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\r\n\x05idMac\x18\x03 \x01(\x0c\"\xb4\x01\n\x15\x43reateDataRoomRequest\x12%\n\x08\x64\x61taRoom\x18\x01 \x01(\x0b\x32\x13.data_room.DataRoom\x12$\n\x17highLevelRepresentation\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x12\x1d\n\x10\x64\x61taRoomMetadata\x18\x03 \x01(\x0cH\x01\x88\x01\x01\x42\x1a\n\x18_highLevelRepresentationB\x13\n\x11_dataRoomMetadata\"\x8c\x01\n\x16\x43reateDataRoomResponse\x12\x14\n\ndataRoomId\x18\x01 \x01(\x0cH\x00\x12?\n\x17\x64\x61taRoomValidationError\x18\x02 \x01(\x0b\x32\x1c.gcg.DataRoomValidationErrorH\x00\x42\x1b\n\x19\x63reate_data_room_response\"\xc6\x02\n\x17\x44\x61taRoomValidationError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x1c\n\x0fpermissionIndex\x18\x02 \x01(\x04H\x00\x88\x01\x01\x12\x1a\n\rcomputeNodeId\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x1d\n\x10userPermissionId\x18\x04 \x01(\tH\x02\x88\x01\x01\x12\'\n\x1a\x61ttestationSpecificationId\x18\x05 \x01(\tH\x03\x88\x01\x01\x12#\n\x16\x61uthenticationMethodId\x18\x06 \x01(\tH\x04\x88\x01\x01\x42\x12\n\x10_permissionIndexB\x10\n\x0e_computeNodeIdB\x13\n\x11_userPermissionIdB\x1d\n\x1b_attestationSpecificationIdB\x19\n\x17_authenticationMethodId\"\x80\x01\n\x1fPublishDatasetToDataRoomRequest\x12\x13\n\x0b\x64\x61tasetHash\x18\x01 \x01(\x0c\x12\x12\n\ndataRoomId\x18\x02 \x01(\x0c\x12\x0e\n\x06leafId\x18\x03 \x01(\t\x12\x15\n\rencryptionKey\x18\x04 \x01(\x0c\x12\r\n\x05scope\x18\x05 \x01(\x0c\"\"\n PublishDatasetToDataRoomResponse\"\xe2\x02\n\x15\x45xecuteComputeRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63omputeNodeIds\x18\x02 \x03(\t\x12\x10\n\x08isDryRun\x18\x03 \x01(\x08\x12\r\n\x05scope\x18\x04 \x01(\x0c\x12>\n\nparameters\x18\x05 \x03(\x0b\x32*.gcg.ExecuteComputeRequest.ParametersEntry\x12\x42\n\x0ctestDatasets\x18\x06 \x03(\x0b\x32,.gcg.ExecuteComputeRequest.TestDatasetsEntry\x1a\x31\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x45\n\x11TestDatasetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.gcg.TestDataset:\x02\x38\x01\"\x8e\x03\n ExecuteDevelopmentComputeRequest\x12\x1d\n\x15\x63onfigurationCommitId\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63omputeNodeIds\x18\x02 \x03(\t\x12\x10\n\x08isDryRun\x18\x03 \x01(\x08\x12\r\n\x05scope\x18\x04 \x01(\x0c\x12I\n\nparameters\x18\x05 \x03(\x0b\x32\x35.gcg.ExecuteDevelopmentComputeRequest.ParametersEntry\x12M\n\x0ctestDatasets\x18\x06 \x03(\x0b\x32\x37.gcg.ExecuteDevelopmentComputeRequest.TestDatasetsEntry\x1a\x31\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x45\n\x11TestDatasetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.gcg.TestDataset:\x02\x38\x01\":\n\x0bTestDataset\x12\x15\n\rencryptionKey\x18\x01 \x01(\x0c\x12\x14\n\x0cmanifestHash\x18\x02 \x01(\x0c\"\'\n\x16\x45xecuteComputeResponse\x12\r\n\x05jobId\x18\x01 \x01(\x0c\"!\n\x10JobStatusRequest\x12\r\n\x05jobId\x18\x01 \x01(\x0c\"3\n\x11JobStatusResponse\x12\x1e\n\x16\x63ompleteComputeNodeIds\x18\x01 \x03(\t\"9\n\x11GetResultsRequest\x12\r\n\x05jobId\x18\x01 \x01(\x0c\x12\x15\n\rcomputeNodeId\x18\x02 \x01(\t\"=\n\x15GetResultsSizeRequest\x12\r\n\x05jobId\x18\x01 \x01(\x0c\x12\x15\n\rcomputeNodeId\x18\x02 \x01(\t\"\'\n\x17GetResultsResponseChunk\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x1a\n\x18GetResultsResponseFooter\"-\n\x17RetrieveDataRoomRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"\xb4\x01\n\x18RetrieveDataRoomResponse\x12%\n\x08\x64\x61taRoom\x18\x01 \x01(\x0b\x32\x13.data_room.DataRoom\x12/\n\x07\x63ommits\x18\x02 \x03(\x0b\x32\x1e.data_room.ConfigurationCommit\x12$\n\x17highLevelRepresentation\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x42\x1a\n\x18_highLevelRepresentation\"-\n\x17RetrieveAuditLogRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"\'\n\x18RetrieveAuditLogResponse\x12\x0b\n\x03log\x18\x01 \x01(\x0c\"3\n\x1dRetrieveDataRoomStatusRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"E\n\x1eRetrieveDataRoomStatusResponse\x12#\n\x06status\x18\x01 \x01(\x0e\x32\x13.gcg.DataRoomStatus\"V\n\x1bUpdateDataRoomStatusRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\x12#\n\x06status\x18\x02 \x01(\x0e\x32\x13.gcg.DataRoomStatus\"\x1e\n\x1cUpdateDataRoomStatusResponse\"6\n RetrievePublishedDatasetsRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"X\n\x10PublishedDataset\x12\x0e\n\x06leafId\x18\x01 \x01(\t\x12\x0c\n\x04user\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x13\n\x0b\x64\x61tasetHash\x18\x04 \x01(\x0c\"U\n!RetrievePublishedDatasetsResponse\x12\x30\n\x11publishedDatasets\x18\x01 \x03(\x0b\x32\x15.gcg.PublishedDataset\"C\n\x1dRemovePublishedDatasetRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\x12\x0e\n\x06leafId\x18\x02 \x01(\t\" \n\x1eRemovePublishedDatasetResponse\"\x94\x01\n CreateConfigurationCommitRequest\x12.\n\x06\x63ommit\x18\x01 \x01(\x0b\x32\x1e.data_room.ConfigurationCommit\x12$\n\x17highLevelRepresentation\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x42\x1a\n\x18_highLevelRepresentation\"5\n!CreateConfigurationCommitResponse\x12\x10\n\x08\x63ommitId\x18\x01 \x01(\x0c\"9\n%GenerateMergeApprovalSignatureRequest\x12\x10\n\x08\x63ommitId\x18\x01 \x01(\x0c\";\n&GenerateMergeApprovalSignatureResponse\x12\x11\n\tsignature\x18\x01 \x01(\x0c\"\xa0\x02\n\x1fMergeConfigurationCommitRequest\x12\x10\n\x08\x63ommitId\x18\x01 \x01(\x0c\x12X\n\x12\x61pprovalSignatures\x18\x02 \x03(\x0b\x32<.gcg.MergeConfigurationCommitRequest.ApprovalSignaturesEntry\x12/\n\"newDataRoomHighLevelRepresentation\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x1a\x39\n\x17\x41pprovalSignaturesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42%\n#_newDataRoomHighLevelRepresentation\"A\n+RetrieveCurrentDataRoomConfigurationRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"t\n,RetrieveCurrentDataRoomConfigurationResponse\x12\x37\n\rconfiguration\x18\x01 \x01(\x0b\x32 .data_room.DataRoomConfiguration\x12\x0b\n\x03pin\x18\x02 \x01(\x0c\"?\n+RetrieveConfigurationCommitApproversRequest\x12\x10\n\x08\x63ommitId\x18\x01 \x01(\x0c\"A\n,RetrieveConfigurationCommitApproversResponse\x12\x11\n\tapprovers\x18\x01 \x03(\t\"6\n\"RetrieveConfigurationCommitRequest\x12\x10\n\x08\x63ommitId\x18\x01 \x01(\x0c\"\x97\x01\n#RetrieveConfigurationCommitResponse\x12.\n\x06\x63ommit\x18\x01 \x01(\x0b\x32\x1e.data_room.ConfigurationCommit\x12$\n\x17highLevelRepresentation\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x42\x1a\n\x18_highLevelRepresentation\"[\n\x18\x43\x61sAuxiliaryStateRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\x12\r\n\x05index\x18\x02 \x01(\x04\x12\x12\n\x05value\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x42\x08\n\x06_value\"Y\n\x19\x43\x61sAuxiliaryStateResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\r\n\x05index\x18\x02 \x01(\x04\x12\x12\n\x05value\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x42\x08\n\x06_value\"/\n\x19ReadAuxiliaryStateRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"F\n\x1aReadAuxiliaryStateResponse\x12(\n\x06values\x18\x02 \x03(\x0b\x32\x18.gcg.AuxiliaryStateValue\"A\n\x13\x41uxiliaryStateValue\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x04\x12\r\n\x05value\x18\x03 \x01(\x0c\"\"\n MergeConfigurationCommitResponse\"|\n\x10\x44riverTaskConfig\x12\x1f\n\x04noop\x18\x01 \x01(\x0b\x32\x0f.gcg.NoopConfigH\x00\x12\x31\n\rstaticContent\x18\x02 \x01(\x0b\x32\x18.gcg.StaticContentConfigH\x00\x42\x14\n\x12\x64river_task_config\"\x0c\n\nNoopConfig\"&\n\x13StaticContentConfig\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\"5\n\x1fRetrieveUsedAirlockQuotaRequest\x12\x12\n\ndataRoomId\x18\x01 \x01(\x0c\"P\n RetrieveUsedAirlockQuotaResponse\x12,\n\rairlockQuotas\x18\x01 \x03(\x0b\x32\x15.gcg.AirlockQuotaInfo\"U\n\x10\x41irlockQuotaInfo\x12\x15\n\rairlockNodeId\x18\x01 \x01(\t\x12\x12\n\nquotaBytes\x18\x02 \x01(\x04\x12\x16\n\x0eusedQuotaBytes\x18\x03 \x01(\x04\"+\n\x16GetResultsSizeResponse\x12\x11\n\tsizeBytes\x18\x01 \x01(\x04*)\n\x0e\x44\x61taRoomStatus\x12\n\n\x06\x41\x63tive\x10\x00\x12\x0b\n\x07Stopped\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gcg_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXECUTECOMPUTEREQUEST_PARAMETERSENTRY._options = None
  _EXECUTECOMPUTEREQUEST_PARAMETERSENTRY._serialized_options = b'8\001'
  _EXECUTECOMPUTEREQUEST_TESTDATASETSENTRY._options = None
  _EXECUTECOMPUTEREQUEST_TESTDATASETSENTRY._serialized_options = b'8\001'
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_PARAMETERSENTRY._options = None
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_PARAMETERSENTRY._serialized_options = b'8\001'
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_TESTDATASETSENTRY._options = None
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_TESTDATASETSENTRY._serialized_options = b'8\001'
  _MERGECONFIGURATIONCOMMITREQUEST_APPROVALSIGNATURESENTRY._options = None
  _MERGECONFIGURATIONCOMMITREQUEST_APPROVALSIGNATURESENTRY._serialized_options = b'8\001'
  _DATAROOMSTATUS._serialized_start=8712
  _DATAROOMSTATUS._serialized_end=8753
  _GCGREQUEST._serialized_start=64
  _GCGREQUEST._serialized_end=1858
  _GCGRESPONSE._serialized_start=1861
  _GCGRESPONSE._serialized_end=3681
  _USERAUTH._serialized_start=3683
  _USERAUTH._serialized_end=3788
  _PKI._serialized_start=3790
  _PKI._serialized_end=3851
  _CREATEDATAROOMREQUEST._serialized_start=3854
  _CREATEDATAROOMREQUEST._serialized_end=4034
  _CREATEDATAROOMRESPONSE._serialized_start=4037
  _CREATEDATAROOMRESPONSE._serialized_end=4177
  _DATAROOMVALIDATIONERROR._serialized_start=4180
  _DATAROOMVALIDATIONERROR._serialized_end=4506
  _PUBLISHDATASETTODATAROOMREQUEST._serialized_start=4509
  _PUBLISHDATASETTODATAROOMREQUEST._serialized_end=4637
  _PUBLISHDATASETTODATAROOMRESPONSE._serialized_start=4639
  _PUBLISHDATASETTODATAROOMRESPONSE._serialized_end=4673
  _EXECUTECOMPUTEREQUEST._serialized_start=4676
  _EXECUTECOMPUTEREQUEST._serialized_end=5030
  _EXECUTECOMPUTEREQUEST_PARAMETERSENTRY._serialized_start=4910
  _EXECUTECOMPUTEREQUEST_PARAMETERSENTRY._serialized_end=4959
  _EXECUTECOMPUTEREQUEST_TESTDATASETSENTRY._serialized_start=4961
  _EXECUTECOMPUTEREQUEST_TESTDATASETSENTRY._serialized_end=5030
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST._serialized_start=5033
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST._serialized_end=5431
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_PARAMETERSENTRY._serialized_start=4910
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_PARAMETERSENTRY._serialized_end=4959
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_TESTDATASETSENTRY._serialized_start=4961
  _EXECUTEDEVELOPMENTCOMPUTEREQUEST_TESTDATASETSENTRY._serialized_end=5030
  _TESTDATASET._serialized_start=5433
  _TESTDATASET._serialized_end=5491
  _EXECUTECOMPUTERESPONSE._serialized_start=5493
  _EXECUTECOMPUTERESPONSE._serialized_end=5532
  _JOBSTATUSREQUEST._serialized_start=5534
  _JOBSTATUSREQUEST._serialized_end=5567
  _JOBSTATUSRESPONSE._serialized_start=5569
  _JOBSTATUSRESPONSE._serialized_end=5620
  _GETRESULTSREQUEST._serialized_start=5622
  _GETRESULTSREQUEST._serialized_end=5679
  _GETRESULTSSIZEREQUEST._serialized_start=5681
  _GETRESULTSSIZEREQUEST._serialized_end=5742
  _GETRESULTSRESPONSECHUNK._serialized_start=5744
  _GETRESULTSRESPONSECHUNK._serialized_end=5783
  _GETRESULTSRESPONSEFOOTER._serialized_start=5785
  _GETRESULTSRESPONSEFOOTER._serialized_end=5811
  _RETRIEVEDATAROOMREQUEST._serialized_start=5813
  _RETRIEVEDATAROOMREQUEST._serialized_end=5858
  _RETRIEVEDATAROOMRESPONSE._serialized_start=5861
  _RETRIEVEDATAROOMRESPONSE._serialized_end=6041
  _RETRIEVEAUDITLOGREQUEST._serialized_start=6043
  _RETRIEVEAUDITLOGREQUEST._serialized_end=6088
  _RETRIEVEAUDITLOGRESPONSE._serialized_start=6090
  _RETRIEVEAUDITLOGRESPONSE._serialized_end=6129
  _RETRIEVEDATAROOMSTATUSREQUEST._serialized_start=6131
  _RETRIEVEDATAROOMSTATUSREQUEST._serialized_end=6182
  _RETRIEVEDATAROOMSTATUSRESPONSE._serialized_start=6184
  _RETRIEVEDATAROOMSTATUSRESPONSE._serialized_end=6253
  _UPDATEDATAROOMSTATUSREQUEST._serialized_start=6255
  _UPDATEDATAROOMSTATUSREQUEST._serialized_end=6341
  _UPDATEDATAROOMSTATUSRESPONSE._serialized_start=6343
  _UPDATEDATAROOMSTATUSRESPONSE._serialized_end=6373
  _RETRIEVEPUBLISHEDDATASETSREQUEST._serialized_start=6375
  _RETRIEVEPUBLISHEDDATASETSREQUEST._serialized_end=6429
  _PUBLISHEDDATASET._serialized_start=6431
  _PUBLISHEDDATASET._serialized_end=6519
  _RETRIEVEPUBLISHEDDATASETSRESPONSE._serialized_start=6521
  _RETRIEVEPUBLISHEDDATASETSRESPONSE._serialized_end=6606
  _REMOVEPUBLISHEDDATASETREQUEST._serialized_start=6608
  _REMOVEPUBLISHEDDATASETREQUEST._serialized_end=6675
  _REMOVEPUBLISHEDDATASETRESPONSE._serialized_start=6677
  _REMOVEPUBLISHEDDATASETRESPONSE._serialized_end=6709
  _CREATECONFIGURATIONCOMMITREQUEST._serialized_start=6712
  _CREATECONFIGURATIONCOMMITREQUEST._serialized_end=6860
  _CREATECONFIGURATIONCOMMITRESPONSE._serialized_start=6862
  _CREATECONFIGURATIONCOMMITRESPONSE._serialized_end=6915
  _GENERATEMERGEAPPROVALSIGNATUREREQUEST._serialized_start=6917
  _GENERATEMERGEAPPROVALSIGNATUREREQUEST._serialized_end=6974
  _GENERATEMERGEAPPROVALSIGNATURERESPONSE._serialized_start=6976
  _GENERATEMERGEAPPROVALSIGNATURERESPONSE._serialized_end=7035
  _MERGECONFIGURATIONCOMMITREQUEST._serialized_start=7038
  _MERGECONFIGURATIONCOMMITREQUEST._serialized_end=7326
  _MERGECONFIGURATIONCOMMITREQUEST_APPROVALSIGNATURESENTRY._serialized_start=7230
  _MERGECONFIGURATIONCOMMITREQUEST_APPROVALSIGNATURESENTRY._serialized_end=7287
  _RETRIEVECURRENTDATAROOMCONFIGURATIONREQUEST._serialized_start=7328
  _RETRIEVECURRENTDATAROOMCONFIGURATIONREQUEST._serialized_end=7393
  _RETRIEVECURRENTDATAROOMCONFIGURATIONRESPONSE._serialized_start=7395
  _RETRIEVECURRENTDATAROOMCONFIGURATIONRESPONSE._serialized_end=7511
  _RETRIEVECONFIGURATIONCOMMITAPPROVERSREQUEST._serialized_start=7513
  _RETRIEVECONFIGURATIONCOMMITAPPROVERSREQUEST._serialized_end=7576
  _RETRIEVECONFIGURATIONCOMMITAPPROVERSRESPONSE._serialized_start=7578
  _RETRIEVECONFIGURATIONCOMMITAPPROVERSRESPONSE._serialized_end=7643
  _RETRIEVECONFIGURATIONCOMMITREQUEST._serialized_start=7645
  _RETRIEVECONFIGURATIONCOMMITREQUEST._serialized_end=7699
  _RETRIEVECONFIGURATIONCOMMITRESPONSE._serialized_start=7702
  _RETRIEVECONFIGURATIONCOMMITRESPONSE._serialized_end=7853
  _CASAUXILIARYSTATEREQUEST._serialized_start=7855
  _CASAUXILIARYSTATEREQUEST._serialized_end=7946
  _CASAUXILIARYSTATERESPONSE._serialized_start=7948
  _CASAUXILIARYSTATERESPONSE._serialized_end=8037
  _READAUXILIARYSTATEREQUEST._serialized_start=8039
  _READAUXILIARYSTATEREQUEST._serialized_end=8086
  _READAUXILIARYSTATERESPONSE._serialized_start=8088
  _READAUXILIARYSTATERESPONSE._serialized_end=8158
  _AUXILIARYSTATEVALUE._serialized_start=8160
  _AUXILIARYSTATEVALUE._serialized_end=8225
  _MERGECONFIGURATIONCOMMITRESPONSE._serialized_start=8227
  _MERGECONFIGURATIONCOMMITRESPONSE._serialized_end=8261
  _DRIVERTASKCONFIG._serialized_start=8263
  _DRIVERTASKCONFIG._serialized_end=8387
  _NOOPCONFIG._serialized_start=8389
  _NOOPCONFIG._serialized_end=8401
  _STATICCONTENTCONFIG._serialized_start=8403
  _STATICCONTENTCONFIG._serialized_end=8441
  _RETRIEVEUSEDAIRLOCKQUOTAREQUEST._serialized_start=8443
  _RETRIEVEUSEDAIRLOCKQUOTAREQUEST._serialized_end=8496
  _RETRIEVEUSEDAIRLOCKQUOTARESPONSE._serialized_start=8498
  _RETRIEVEUSEDAIRLOCKQUOTARESPONSE._serialized_end=8578
  _AIRLOCKQUOTAINFO._serialized_start=8580
  _AIRLOCKQUOTAINFO._serialized_end=8665
  _GETRESULTSSIZERESPONSE._serialized_start=8667
  _GETRESULTSSIZERESPONSE._serialized_end=8710
# @@protoc_insertion_point(module_scope)



from google.protobuf.json_format import MessageToDict
from ..proto import serialize_length_delimited, ComputeNodeFormat, parse_length_delimited
from ..node import Node
from .proto import (
    AzureBlobStorageWorkerConfiguration,
    SinkInput,
    ExportRole,
)

class DataSourceAzureBlobStorage(Node):
    """
    Compute node that fetches a dataset from a Azure blob storage.
    """

    def __init__(
            self,
            name: str,
            credentials_dependency: str,
    ) -> None:
        config = AzureBlobStorageWorkerConfiguration(
            credentialsDependency=credentials_dependency,
            importRole={}
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.azure-blob-storage-worker",
            dependencies=[credentials_dependency],
            output_format=ComputeNodeFormat.RAW
        )

class DataSinkAzureBlobStorage(Node):
    """
    Compute node that exports data to Azure blob storage.
    """

    def __init__(
            self,
            name: str,
            credentials_dependency: str,
            input: SinkInput,
    ) -> None:
        config = AzureBlobStorageWorkerConfiguration(
            credentialsDependency=credentials_dependency,
            exportRole=ExportRole(input=input)
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.azure-blob-storage-worker",
            dependencies=[credentials_dependency, input.dependency],
            output_format=ComputeNodeFormat.RAW
        )

class AzureBlobStorageWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = AzureBlobStorageWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)

__all__ = [
    "DataSourceAzureBlobStorage",
    "DataSinkAzureBlobStorage",
    "AzureBlobStorageWorkerDecoder"
]



from __future__ import annotations
import uuid
import base64
import json
from typing import List, Any, Optional, Dict, Tuple

from .proto import (
    AuthenticationMethod,
    UserPermission,
    ComputeNode,
    ComputeNodeLeaf,
    ComputeNodeBranch,
    Permission,
    DataRoom,
    ComputeNodeParameter,
    AttestationSpecification,
    StaticDataRoomPolicy,
    AffectedDataOwnersApprovePolicy,
    ConfigurationModification,
    AddModification,
    ChangeModification,
    ConfigurationElement,
    ConfigurationCommit,
    DataRoomConfiguration,
    ComputeNodeProtocol,
    ComputeNodeAirlock,
)
from .proto import (
    GovernanceProtocol as GovernanceProtocolProto,
    serialize_length_delimited,
)
from .node import Node
from .permission import Permissions
from .types import (
    EnclaveSpecification,
    MatchingId,
)
from .client import Client
from .keychain import Keychain
from decentriq_dcr_compiler import compiler
from decentriq_dcr_compiler.schemas.data_lab import (
    DataLab,
)
from .data_lab import DataLab, DataLabConfig, ExistingDataLab
from .lookalike_media import LookalikeMediaDcr, ExistingLookalikeMediaDcr
from .lookup_tables import MATCHING_ID_INTERNAL_LOOKUP

from .helpers import (
    create_session_from_driver_spec,
    get_latest_enclave_specs_as_dictionary,
)


__all__ = [
    "DataRoomBuilder",
    "DataRoomCommitBuilder",
    "DataRoomModificationsBuilder",
    "GovernanceProtocol",
]


def _extract_configuration_elements(
    config: Optional[DataRoomConfiguration],
) -> Tuple[
    List[Tuple[str, AttestationSpecification]],
    List[Tuple[str, AuthenticationMethod]],
    List[Tuple[str, UserPermission]],
    List[Tuple[str, ComputeNode]],
]:
    attestation_specs = {}
    authentication_methods = {}
    user_permissions = {}
    compute_nodes = {}

    items = config.elements if config else []

    for element in items:
        if element.HasField("attestationSpecification"):
            attestation_specs[element.id] = element.attestationSpecification
        elif element.HasField("authenticationMethod"):
            authentication_methods[element.id] = element.authenticationMethod
        elif element.HasField("userPermission"):
            user_permissions[element.id] = element.userPermission
        elif element.HasField("computeNode"):
            compute_nodes[element.id] = element.computeNode

    return (
        list(attestation_specs.items()),
        list(authentication_methods.items()),
        list(user_permissions.items()),
        list(compute_nodes.items()),
    )


def _find_matching_proto_object(
    proto_object: Any, other_proto_objects: List[Tuple[str, Any]]
) -> Optional[str]:
    existing_node_id = next(
        (
            node_id
            for node_id, other_proto_object in other_proto_objects
            if proto_object == other_proto_object
        ),
        None,
    )
    return existing_node_id


class GovernanceProtocol:
    """
    The protocol that defines whether and how a data room can be changed
    after it has been published.
    """

    @staticmethod
    def static():
        """
        The data room cannot be changed after it has been published.
        Participants are still allowed to execute development computations
        as long as the required permissions have been granted.
        """
        return GovernanceProtocolProto(staticDataRoomPolicy=StaticDataRoomPolicy())

    @staticmethod
    def affected_data_owners_approve():
        """
        The addition of compute nodes must be approved by all data owners
        on whose data the new compute node will depend.
        """
        return GovernanceProtocolProto(
            affectedDataOwnersApprovePolicy=AffectedDataOwnersApprovePolicy()
        )


def _deduplicate_proto_objects(protos):
    """
    Helper function for deduplicating a list of proto messages while
    keeping their original order.
    """
    # Python dicts for Python >= 3.7 keep insertion order
    protos_map = {proto.SerializeToString(): proto for proto in protos}
    return [proto for proto in protos_map.values()]


def _extend_repeated_field(repeated_field, new_protos) -> bool:
    """
    Helper function for extending a repeated field, making sure
    not to insert any duplicates while keeping insertion order.
    """
    repeated_field_map = {proto.SerializeToString(): proto for proto in repeated_field}
    did_append_something = False
    for proto in new_protos:
        proto_key = proto.SerializeToString()
        if proto_key not in repeated_field_map:
            repeated_field_map[proto_key] = proto
            repeated_field.append(proto)
            did_append_something = True
    return did_append_something


class DataRoomBuilder:
    """
    A helper class to ease the building process of a data clean room.
    """

    name: str
    governance_protocol: GovernanceProtocolProto
    description: Optional[str]
    dcr_secret_id: Optional[bytes]
    modifications_builder: DataRoomModificationsBuilder

    def __init__(
        self,
        name: str,
        enclave_specs: Dict[str, EnclaveSpecification],
        governance_protocol: GovernanceProtocolProto = GovernanceProtocol.static(),
        *,
        add_basic_user_permissions: bool = True,
        description: str = None,
        dcr_secret_id: bytes = None,
    ) -> None:
        """
        Create a data room builder object.

        **Parameters**:
        - `name`: The name of the data room to be created.
        - `enclave_specs`: The enclave specification set in which to look-up
            enclave specs for enclaves responsible for executing compute nodes.
            These specs are provided by the `decentriq_platform.enclave_specifications`
            catalogue.
        - `governance_protocol`: The protocol that defines whether and how a
            data room can be changed after it has been published.
        - `add_basic_user_permissions`: Whether to add basic user permissions
            for each participant. These are:
            1. Permission to retrieve the data room definition
            2. Permission to retrieve the status of the data room
            3. Permission to retrieve the audit log
            4. Permission to retrieve the list of datasets that have been published to the data room
            5. Permission to run development computations
        - `description`: Description of the data room.
        """
        assert name, "The DCR must have a non-empty name"

        self.modifications_builder = DataRoomModificationsBuilder(
            enclave_specs, add_basic_user_permissions=add_basic_user_permissions
        )
        self.name = name
        self.description = description
        self.governance_protocol = governance_protocol
        self.enclave_specs = enclave_specs
        self.dcr_secret_id = dcr_secret_id

        if description:
            self.add_description(description)

    def add_data_node(
        self, name: str, is_required: bool = False, node_id: Optional[str] = None
    ) -> str:
        """
        Add a new data node. If the node is marked as required, any computation
        which includes it as a dependency will not start in case no data has
        been provided yet.

        **Parameters**:
        - `name`: Name of the data node.
        - `is_required`: If true, any computation which depends on this data node
            can only be run if data has been provided for this node.
        - `node_id`: A custom identifier for this node.
            If not specified, the identifier is generated automatically.

        **Returns**:
        The id that was assigned to the added data node.
        This id will be needed when connecting a dataset or when permissions condering
        this node are defined.
        """
        return self.modifications_builder.add_data_node(
            name, is_required=is_required, node_id=node_id
        )

    def add_compute_node(
        self,
        node: Node,
        node_id: Optional[str] = None,
        attestation_id: Optional[str] = None,
    ) -> str:
        """
        Add a new compute node. Specific compute node classes are provided either
        by the main package or by the respective compute submodules.

        **Parameters**:
        - `node`: The node object to add.
        - `node_id`: A customer identifier for this node. If not specified,
            the identifier is generated automatically.
        - `attestation_id`: An identifier for a concrete attestation specification
            to use for this compute node. If not specified, the specification is
            chosen automatically based on the type of compute node.

        **Returns**:
        The id that was assigned to the added compute node.
        This id is needed when running computations or when adding permissions
        concerning this node.
        """
        return self.modifications_builder.add_compute_node(
            node,
            node_id=node_id,
            attestation_id=attestation_id,
        )

    def add_airlock_node(
        self,
        quota_bytes: int,
        airlocked_dependency: str,
        node_name: str = "",
        node_id: Optional[str] = None,
    ) -> str:
        """
        Add a new airlock node. Specific compute node classes are provided either
        by the main package or by the respective compute submodules.

        **Parameters**:
        - `quota_bytes`: The maximum quota each participant can use in bytes.
        - `airlocked_dependencies`: the id of the branch or leaf nodes that this airlock can pull data
          from.
        - `node_id`: A custom identifier for this node. If not specified,
            the identifier is generated automatically.
        **Returns**:
        The id that was assigned to the added compute node.
        This id is needed when running computations or when adding permissions
        concerning this node.
        """
        return self.modifications_builder.add_airlock_node(
            quota_bytes,
            airlocked_dependency,
            node_name,
            node_id,
        )

    def add_user_permission(
        self,
        email: str,
        authentication_method: AuthenticationMethod,
        permissions: List[Permission],
    ):
        """
        Add permissions for a user.
        The authentication is performed on the enclave side based on the method supplied.
        """
        return self.modifications_builder.add_user_permission(
            email, authentication_method, permissions
        )

    def add_description(self, description: str):
        """Add a description to the data room being built."""
        self.description = description

    def build(self) -> DataRoom:
        """
        Finalize data room contruction.

        Built data rooms still need to be published by a
        `decentriq_platform.Session` before they can be interacted with.

        **Returns**:
        A tuple containing the following elements:
        (1) A data room object that stores the core properties of the DCR, such
        as its name, its owner, and the government protocol that defines what changes
        can be made to the data room's configuration.
        (2) A list of the recorded modifications that will be applied to the initially
        empty data room configuration within the enclave.
        """
        data_room = DataRoom()
        data_room.name = self.name
        if self.description:
            data_room.description = self.description
        data_room.id = DataRoomBuilder._generate_id()
        data_room.governanceProtocol.CopyFrom(self.governance_protocol)
        data_room.initialConfiguration.CopyFrom(self.modifications_builder.build_flat())

        return data_room

    @staticmethod
    def _generate_id():
        return str(uuid.uuid4())


class DataRoomModificationsBuilder:
    """
    Builder class for constructing lists of modifications that can be
    applied to existing data room configurations.
    """

    data_room_id: Optional[str]
    history_pin: Optional[str]

    new_attestation_specs: List[Tuple[str, AttestationSpecification]]
    new_authentication_methods: List[Tuple[str, AuthenticationMethod]]
    new_user_permissions: List[Tuple[str, UserPermission]]
    new_compute_nodes: List[Tuple[str, ComputeNode]]

    existing_attestation_specs: List[Tuple[str, AttestationSpecification]]
    existing_authentication_methods: List[Tuple[str, AuthenticationMethod]]
    existing_user_permissions: List[Tuple[str, UserPermission]]

    change_user_permissions: List[Tuple[str, UserPermission]]
    change_attestation_specs: List[Tuple[str, AttestationSpecification]]

    enclave_specs: Dict[str, EnclaveSpecification]
    add_basic_user_permissions: bool

    def __init__(
        self,
        enclave_specs: Dict[str, EnclaveSpecification],
        *,
        data_room_id: Optional[str] = None,
        data_room_configuration: Optional[DataRoomConfiguration] = None,
        history_pin: Optional[str] = None,
        add_basic_user_permissions: bool = True,
    ) -> None:
        """
        Construct a builder object for constructing lists of modifications that
        can be applied to existing data room configurations.

        **Parameters**:
        - `enclave_specs`: The enclave specification set in which to lookup
            enclave specs for enclaves responsible for executing compute nodes.
            These specs are provided by the `decentriq_platform.enclave_specifications`
            catalogue.
        - `data_room_id`: The data room to which the modifications should be
            applied.
        - `data_room_configuration`: The current configuration of the data room
            to be altered.
        - `history_pin`: The current history pin that identifies a specific
            point in time within a line of configuration changes.
        - `add_basic_user_permissions`: Whether to add basic user permissions
            for each participant. These are:
            1. Permission to retrieve the data room definition
            2. Permission to retrieve the status of the data room
            3. Permission to retrieve the audit log
            4. Permission to retrieve the list of datasets that have been published to the data room
            5. Permission to run development computations
        """
        self.data_room_id = data_room_id
        self.history_pin = history_pin
        self.enclave_specs = enclave_specs
        self.add_basic_user_permissions = add_basic_user_permissions

        self.new_attestation_specs = []
        self.new_authentication_methods = []
        self.new_user_permissions = []
        self.new_compute_nodes = []

        self.change_user_permissions = []
        self.change_attestation_specs = []
        self.change_compute_nodes = []

        extracted_conf_elements = _extract_configuration_elements(
            data_room_configuration
        )
        (
            self.existing_attestation_specs,
            self.existing_authentication_methods,
            self.existing_user_permissions,
            self.existing_compute_nodes,
        ) = extracted_conf_elements

    def add_data_node(
        self, name: str, is_required: bool = False, node_id: Optional[str] = None
    ) -> str:
        """
        Add a new data node. If the node is marked as required, any computation
        which includes it as a dependency will not start in case no data has
        been provided yet.

        **Parameters**:
        - `name`: Name of the data node.
        - `is_required`: If true, any computation which depends on this data node
            can only be run if data has been provided for this node.
        - `node_id`: A custom identifier for this node.
            If not specified, the identifier is generated automatically.

        **Returns**:
        The id that was assigned to the added data node.
        This id will be needed when connecting a dataset or when permissions condering
        this node are defined.
        """
        node_id = (
            DataRoomModificationsBuilder._generate_id() if not node_id else node_id
        )
        node = ComputeNode(nodeName=name, leaf=ComputeNodeLeaf(isRequired=is_required))
        self.new_compute_nodes.append((node_id, node))
        return node_id

    def add_parameter_node(
        self, name: str, is_required: bool = False, node_id: Optional[str] = None
    ) -> str:
        """
        Add a new parameter node. If the node is marked as required, any computation
        which includes it as a dependency will not start in case no data has
        been provided yet.

        **Parameters**:
        - `name`: Name of the parameter node.
        - `is_required`: If true, any computation which depends on this parameter node
            can only be run if the compute request also provides data for the parameter.
        - `node_id`: A custom identifier for this node.
            If not specified, the identifier is generated automatically.

        **Returns**:
        The id that was assigned to the added parameter node.
        """
        node_id = (
            DataRoomModificationsBuilder._generate_id() if not node_id else node_id
        )
        bla = ComputeNodeParameter(isRequired=is_required)
        node = ComputeNode(
            nodeName=name, parameter=ComputeNodeParameter(isRequired=is_required)
        )
        self.new_compute_nodes.append((node_id, node))
        return node_id

    def _add_attestation_specification(
        self, attestation_specification: AttestationSpecification
    ) -> str:
        existing_id = _find_matching_proto_object(
            attestation_specification,
            self.existing_attestation_specs + self.new_attestation_specs,
        )
        if not existing_id:
            new_id = DataRoomModificationsBuilder._generate_id()
            self.new_attestation_specs.append((new_id, attestation_specification))
            return new_id
        else:
            return existing_id

    def change_attestation_specification(
        self, attestation_id: str, attestation_specification: AttestationSpecification
    ) -> None:
        """
        Change a particular attestation specification within an existing data room
        configuration to a different one.
        This can be useful if a certain computation is run by a particular worker enclave
        that needs to be upgraded to a newer version.

        **Parameters**:
        - `attestation_id`: The id of the attestation to be replaced.
        - `attestation_specification`: The attestation specification protobuf object.
        """
        existing_spec_ix = None
        for ix, (a_id, attestation) in enumerate(self.new_attestation_specs):
            if a_id == attestation_id:
                existing_spec_ix = ix
                break

        if existing_spec_ix:
            old_spec_id, old_spec = self.new_attestation_specs[existing_spec_ix]
            self.new_attestation_specs[existing_spec_ix] = (
                old_spec_id,
                attestation_specification,
            )
        elif attestation_id in dict(self.existing_attestation_specs):
            self.change_attestation_specs.append(
                (attestation_id, attestation_specification)
            )
        else:
            raise Exception(
                f"There is no attestation specification with id {attestation_id}"
            )

    def add_compute_node(
        self,
        node: Node,
        node_id: Optional[str] = None,
        attestation_id: Optional[str] = None,
    ) -> str:
        """
        Add a new compute node. Specific compute node classes are provided either
        by the main package or by the respective compute submodules.

        **Parameters**:
        - `node`: The node object to add.
        - `node_id`: A customer identifier for this node. If not specified,
            the identifier is generated automatically.
        - `attestation_id`: An identifier for a concrete attestation specification
            to use for this compute node. If not specified, the specification is
            chosen automatically based on the type of compute node.

        **Returns**:
        The id that was assigned to the added compute node.
        This id is needed when running computations or when adding permissions
        concerning this node.
        """
        if self.enclave_specs:
            if node.enclave_type in self.enclave_specs:
                attestation_proto = self.enclave_specs[node.enclave_type]["proto"]
                worker_protocol_version = max(
                    self.enclave_specs[node.enclave_type]["workerProtocols"]
                )
                node_protocol = ComputeNodeProtocol(version=worker_protocol_version)
            else:
                raise Exception(
                    f"This compute node requires an enclave of type '{node.enclave_type}' but no"
                    " corresponding enclave spec was provided to this builder."
                )
        else:
            raise Exception(
                "You need to provide a dictionary of enclave specifications to the"
                " builder before calling this method."
            )

        if attestation_id:
            final_attestation_id = attestation_id
        else:
            final_attestation_id = self._add_attestation_specification(
                attestation_proto
            )

        node_id = (
            DataRoomModificationsBuilder._generate_id() if not node_id else node_id
        )
        proto_node = ComputeNode(
            nodeName=node.name,
            branch=ComputeNodeBranch(
                config=node.config,
                attestationSpecificationId=final_attestation_id,
                dependencies=node.dependencies,
                outputFormat=node.output_format,
                protocol=node_protocol,
            ),
        )
        self.new_compute_nodes.append((node_id, proto_node))
        return node_id

    def add_airlock_node(
        self,
        quota_bytes: int,
        airlocked_dependency: str,
        node_name: str = "",
        node_id: Optional[str] = None,
    ) -> str:
        """
        Add a new airlock node. Specific compute node classes are provided either
        by the main package or by the respective compute submodules.

        **Parameters**:
        - `quota_bytes`: The maximum quota each participant can use in bytes.
        - `airlocked_dependency`: the id of the branch or leaf node that this airlock can pull data
          from.
        - `node_id`: A custom identifier for this node. If not specified,
            the identifier is generated automatically.
        **Returns**:
        The id that was assigned to the added compute node.
        This id is needed when running computations or when adding permissions
        concerning this node.
        """
        node_id = (
            DataRoomModificationsBuilder._generate_id() if not node_id else node_id
        )
        proto_node = ComputeNode(
            nodeName=node_name,
            airlock=ComputeNodeAirlock(
                quotaBytes=quota_bytes,
                airlockedDependency=airlocked_dependency,
            ),
        )
        self.new_compute_nodes.append((node_id, proto_node))
        return node_id

    def change_airlock_node(
        self,
        airlock_node_id: str,
        quota_bytes: int,
    ) -> None:
        """
        **Parameters:**
        - `airlock_node_id`: Id of the airlock node to be changed.
        - `quota_bytes`: New quota for the airlock id.
        """
        existing_airlock_ix = None
        for ix, (c_id, compute_node) in enumerate(self.new_compute_nodes):
            if c_id == airlock_node_id:
                existing_airlock_ix = ix
                break

        if existing_airlock_ix:
            old_node_id, old_node = self.new_compute_nodes[existing_airlock_ix]
            old_node.airlock.quotaBytes = quota_bytes
            self.new_compute_nodes[existing_airlock_ix] = (
                old_node_id,
                old_node,
            )
        else:
            previous_compute_node = dict(self.existing_compute_nodes).get(
                airlock_node_id
            )
            if not previous_compute_node or not previous_compute_node.HasField(
                "airlock"
            ):
                raise Exception(f"There is no airlock node with id {airlock_node_id}")
            changed_compute_node = ComputeNode()
            changed_compute_node.CopyFrom(previous_compute_node)
            changed_compute_node.airlock.quotaBytes = quota_bytes
            self.change_compute_nodes.append((airlock_node_id, changed_compute_node))

    def _add_authentication_method(
        self, authentication_method: AuthenticationMethod
    ) -> str:
        existing_id = _find_matching_proto_object(
            authentication_method,
            self.existing_authentication_methods + self.new_authentication_methods,
        )
        if existing_id:
            return existing_id
        else:
            new_id = DataRoomModificationsBuilder._generate_id()
            self.new_authentication_methods.append((new_id, authentication_method))
            return new_id

    def add_user_permission(
        self,
        email: str,
        authentication_method: AuthenticationMethod,
        permissions: List[Permission],
    ):
        """
        Add permissions for a user.
        The authentication is performed on the enclave side based on the method supplied.
        """
        authentication_method_id = self._add_authentication_method(
            authentication_method
        )

        if self.add_basic_user_permissions:
            permissions.extend(
                [
                    Permissions.retrieve_data_room_status(),
                    Permissions.retrieve_audit_log(),
                    Permissions.retrieve_data_room(),
                    Permissions.retrieve_published_datasets(),
                    Permissions.execute_development_compute(),
                ]
            )

        # Check whether a set of permissions has already been added in a previous
        # call, and extend the permissions in case the authentication method matches.
        # This is required because helper functions might add permissions
        # to the builder before this method is called.
        existing_permission_old = [
            permission
            for permission in self.existing_user_permissions
            if email == permission[1].email
            and authentication_method_id == permission[1].authenticationMethodId
        ]
        existing_permission_new = [
            permission
            for permission in self.new_user_permissions
            if email == permission[1].email
            and authentication_method_id == permission[1].authenticationMethodId
        ]

        if existing_permission_old:
            existing_id, existing_permission = existing_permission_old[0]
            did_append_something = _extend_repeated_field(
                existing_permission.permissions, permissions
            )
            # Do not add a change modification if none is necessary
            if did_append_something:
                self.change_user_permissions.append((existing_id, existing_permission))
        elif existing_permission_new:
            _, existing_permission = existing_permission_new[0]
            _extend_repeated_field(existing_permission.permissions, permissions)
        else:
            permission = UserPermission(
                email=email,
                authenticationMethodId=authentication_method_id,
                permissions=_deduplicate_proto_objects(permissions),
            )
            self.new_user_permissions.append(
                (DataRoomModificationsBuilder._generate_id(), permission)
            )

    def build_flat(self) -> DataRoomConfiguration:
        configuration_elements = []

        for spec_id, attestation_spec in self.new_attestation_specs:
            configuration_elements.append(
                ConfigurationElement(
                    id=spec_id, attestationSpecification=attestation_spec
                )
            )

        for node_id, compute_node in self.new_compute_nodes:
            configuration_elements.append(
                ConfigurationElement(id=node_id, computeNode=compute_node)
            )

        for auth_id, auth_method in self.new_authentication_methods:
            configuration_elements.append(
                ConfigurationElement(id=auth_id, authenticationMethod=auth_method)
            )

        for perm_id, permission in self.new_user_permissions:
            configuration_elements.append(
                ConfigurationElement(id=perm_id, userPermission=permission)
            )

        configuration = DataRoomConfiguration(elements=configuration_elements)
        return configuration

    def build(self) -> List[ConfigurationModification]:
        """
        Build the list of configuration modifications.
        """
        modifications = []

        for spec_id, attestation_spec in self.new_attestation_specs:
            modifications.append(
                ConfigurationModification(
                    add=AddModification(
                        element=ConfigurationElement(
                            id=spec_id, attestationSpecification=attestation_spec
                        )
                    )
                )
            )

        for node_id, compute_node in self.new_compute_nodes:
            modifications.append(
                ConfigurationModification(
                    add=AddModification(
                        element=ConfigurationElement(
                            id=node_id, computeNode=compute_node
                        )
                    )
                )
            )

        for auth_id, auth_method in self.new_authentication_methods:
            modifications.append(
                ConfigurationModification(
                    add=AddModification(
                        element=ConfigurationElement(
                            id=auth_id, authenticationMethod=auth_method
                        )
                    )
                ),
            )

        for perm_id, permission in self.new_user_permissions:
            modifications.append(
                ConfigurationModification(
                    add=AddModification(
                        element=ConfigurationElement(
                            id=perm_id, userPermission=permission
                        )
                    )
                )
            )

        for node_id, compute_node in self.change_compute_nodes:
            modifications.append(
                ConfigurationModification(
                    change=ChangeModification(
                        element=ConfigurationElement(
                            id=node_id,
                            computeNode=compute_node,
                        )
                    )
                )
            )

        for perm_id, permission in self.change_user_permissions:
            modifications.append(
                ConfigurationModification(
                    change=ChangeModification(
                        element=ConfigurationElement(
                            id=perm_id, userPermission=permission
                        )
                    )
                )
            )

        for attestation_id, attestation_spec in self.change_attestation_specs:
            modifications.append(
                ConfigurationModification(
                    change=ChangeModification(
                        element=ConfigurationElement(
                            id=attestation_id, attestationSpecification=attestation_spec
                        )
                    )
                )
            )

        return modifications

    @staticmethod
    def _generate_id():
        return str(uuid.uuid4())


class DataRoomCommitBuilder:
    """
    A helper class to build a data room configuration commit,
    i.e. a list of modifications that are to be applied to the configuration
    of an existing data room.
    """

    name: str
    data_room_id: str
    history_pin: str
    modifications_builder: DataRoomModificationsBuilder

    def __init__(
        self,
        name: str,
        data_room_id: str,
        current_configuration: DataRoomConfiguration,
        history_pin: str,
        enclave_specs: Dict[str, EnclaveSpecification],
        *,
        add_basic_user_permissions: bool = False,
    ):
        """
        Construct a builder object for constructing new data room
        configuration commits.
        A configuraton commit contains a list of modifications that
        can be applied to existing data room configuration.

        **Parameters**:
        - `data_room_id`: The data room to which the modifications should be
            applied.
        - `data_room_configuration`: The current configuration of the data room
            to be altered.
        - `history_pin`: The current history pin that identifies a specific
            point in time within a line of configuration changes.
        - `enclave_specs`: The enclave specification set in which to look-up
            enclave specs for enclaves responsible for executing compute nodes.
            These specs are provided by the `decentriq_platform.enclave_specifications`
            catalogue.
        - `add_basic_user_permissions`: Whether to add basic user permissions
            for each participant. These are:
            1. Permission to retrieve the data room definition
            2. Permission to retrieve the status of the data room
            3. Permission to retrieve the audit log
            4. Permission to retrieve the list of datasets that have been published to the data room
            5. Permission to run development computations
        """
        self.name = name
        self.data_room_id = data_room_id
        self.history_pin = history_pin
        self.modifications_builder = DataRoomModificationsBuilder(
            enclave_specs,
            data_room_id=data_room_id,
            history_pin=history_pin,
            data_room_configuration=current_configuration,
            add_basic_user_permissions=add_basic_user_permissions,
        )

    def add_data_node(
        self, name: str, is_required: bool = False, node_id: Optional[str] = None
    ) -> str:
        """
        Add a new data node. If the node is marked as required, any computation
        which includes it as a dependency will not start in case no data has
        been provided yet.

        **Parameters**:
        - `name`: Name of the data node.
        - `is_required`: If true, any computation which depends on this data node
            can only be run if data has been provided for this node.
        - `node_id`: A custom identifier for this node.
            If not specified, the identifier is generated automatically.

        **Returns**:
        The id that was assigned to the added data node.
        This id will be needed when connecting a dataset or when permissions condering
        this node are defined.
        """
        return self.modifications_builder.add_data_node(
            name, is_required=is_required, node_id=node_id
        )

    def add_parameter_node(
        self, name: str, is_required: bool = False, node_id: Optional[str] = None
    ) -> str:
        """
        Add a new parameter node. If the node is marked as required, any computation
        which includes it as a dependency will not start in case no data has
        been provided yet.

        **Parameters**:
        - `name`: Name of the parameter node.
        - `is_required`: If true, any computation which depends on this parameter node
            can only be run if the compute request also provides data for the parameter.
        - `node_id`: A custom identifier for this node.
            If not specified, the identifier is generated automatically.

        **Returns**:
        The id that was assigned to the added parameter node.
        """
        return self.modifications_builder.add_parameter_node(
            name, is_required=is_required, node_id=node_id
        )

    def change_attestation_specification(
        self, attestation_id: str, attestation_specification: AttestationSpecification
    ) -> None:
        self.modifications_builder.change_attestation_specification(
            attestation_id, attestation_specification
        )

    def add_compute_node(
        self,
        node: Node,
        node_id: Optional[str] = None,
        attestation_id: Optional[str] = None,
    ) -> str:
        """
        Add a new compute node. Specific compute node classes are provided either
        by the main package or by the respective compute submodules.

        **Parameters**:
        - `node`: The node object to add.
        - `node_id`: A customer identifier for this node. If not specified,
            the identifier is generated automatically.
        - `attestation_id`: An identifier for a concrete attestation specification
            to use for this compute node. If not specified, the specification is
            chosen automatically based on the type of compute node.

        **Returns**:
        The id that was assigned to the added compute node.
        This id is needed when running computations or when adding permissions
        concerning this node.
        """
        return self.modifications_builder.add_compute_node(
            node, node_id=node_id, attestation_id=attestation_id
        )

    def add_user_permission(
        self,
        email: str,
        authentication_method: AuthenticationMethod,
        permissions: List[Permission],
    ):
        """
        Add permissions for a user.
        The authentication is performed on the enclave side based on the method supplied.
        """
        return self.modifications_builder.add_user_permission(
            email, authentication_method, permissions
        )

    def add_airlock_node(
        self,
        quota_bytes: int,
        airlocked_dependency: str,
        node_name: str = "",
        node_id: Optional[str] = None,
    ) -> str:
        """
        Add a new airlock node. Specific compute node classes are provided either
        by the main package or by the respective compute submodules.

        **Parameters**:
        - `quota_bytes`: The maximum quota each participant can use in bytes.
        - `airlocked_dependency`: the id of the branch or leaf nodes that this airlock can pull data
          from.
        - `node_id`: A custom identifier for this node. If not specified,
            the identifier is generated automatically.
        **Returns**:
        The id that was assigned to the added compute node.
        This id is needed when running computations or when adding permissions
        concerning this node.
        """
        return self.modifications_builder.add_airlock_node(
            quota_bytes,
            airlocked_dependency,
            node_name,
            node_id,
        )

    def change_airlock_node(
        self,
        airlock_node_id: str,
        quota_bytes: int,
    ) -> None:
        return self.modifications_builder.change_airlock_node(
            airlock_node_id, quota_bytes
        )

    @staticmethod
    def _generate_id():
        return str(uuid.uuid4())

    def build(self):
        """
        Build the data room configuration commit.

        The built commit still needs to be published and merged in order for
        it to be made part of the data room configuration.
        """
        return ConfigurationCommit(
            id=DataRoomCommitBuilder._generate_id(),
            name=self.name,
            dataRoomId=bytes.fromhex(self.data_room_id),
            dataRoomHistoryPin=bytes.fromhex(self.history_pin),
            modifications=self.modifications_builder.build(),
        )


class DataLabBuilder:
    """
    A helper class to build a Data Lab.
    """

    def __init__(
        self,
        client: Client,
    ):
        self.name = None
        self.has_demographics = False
        self.has_embeddings = False
        self.num_embeddings = 0
        self.matching_id = MatchingId.STRING
        self.validation_id = None
        self.client = client
        self.existing = False
        self.data_lab_id = None

    def with_name(self, name: str):
        """
        Set the name of the DataLab.

        **Parameters**:
        - `name`: Name to be used for the DataLab.
        """
        self.name = name

    def with_matching_id_format(self, matching_id: MatchingId):
        """
        Set the matching ID format.

        **Parameters**:
        - `matching_id`: The type of matching ID to use.
        """
        self.matching_id = matching_id

    def with_demographics(self):
        """
        Enable demographics in the DataLab.
        """
        self.has_demographics = True

    def with_embeddings(self, num_embeddings: int):
        """
        Enable embeddings in the DataLab.

        **Parameters**:
        - `num_embeddings`: The number of embeddings the DataLab should use.
        """
        self.has_embeddings = True
        self.num_embeddings = num_embeddings

    def from_existing(self, data_lab_id: str, keychain: Keychain):
        """
        Construct a new DataLab from an existing DataLab with the given ID.

        **Parameters**:
        - `data_lab_id`: The ID of the existing DataLab.
        - `keychain`: The keychain to use to provision datasets from the old DataLab to the new DataLab.
        """
        self.existing = True
        self.data_lab_id = data_lab_id
        self.keychain = keychain

    def build(self) -> DataLab:
        """
        Build the DataLab.
        """
        if self.existing:
            # Build a new DataLab from an existing one.
            # The new DataLab will have the same configuration as the existing one.
            data_lab_definition = self.client.get_data_lab(self.data_lab_id)
            cfg = DataLabConfig(
                data_lab_definition["name"],
                data_lab_definition["requireDemographicsDataset"],
                data_lab_definition["requireEmbeddingsDataset"],
                data_lab_definition["numEmbeddings"],
                data_lab_definition["matchingIdFormat"],
            )
            existing_data_lab = ExistingDataLab(data_lab_definition, self.keychain)
            return DataLab(self.client, cfg, existing_data_lab)
        else:
            # Build a new DataLab using the specified enclave specifications.
            cfg = DataLabConfig(
                self.name,
                self.has_demographics,
                self.has_embeddings,
                self.num_embeddings,
                self.matching_id,
            )
            return DataLab(self.client, cfg)


class LookalikeMediaDcrBuilder:
    """
    A helper class to build a Lookalike Media DCR
    """

    def __init__(
        self,
        client: Client,
    ) -> None:
        self.client = client
        self.name = None
        self.main_publisher_email = None
        self.publisher_emails = []
        self.main_advertiser_email = None
        self.advertiser_emails = []
        self.matching_id = None
        self.audit_log_retrieval = True
        self.dev_computations = True
        self.agency_emails = []
        self.observer_emails = []
        self.existing = False

    def with_name(self, name: str):
        """
        Set the name of the Lookalike Media DCR.
        This is required when creating **new** Lookalike Media DCRs only.
        When creating from an existing DCR, the existing name will be used.

        **Parameters**:
        - `name`: Name to be used for the Lookalike Media DCR.
        """
        self.name = name

    def with_publisher_emails(self, main: str, additional: Optional[List[str]] = None):
        """
        Set the publisher email addresses.

        **Parameters**:
        - `main`: The main publisher email address.
        - `additional`: Optional list of additional publisher email addresses.
        """
        self.main_publisher_email = main
        if additional is not None:
            self.publisher_emails = additional

    def with_advertiser_emails(self, main: str, additional: Optional[List[str]] = None):
        """
        Set the advertiser email addresses.

        **Parameters**:
        - `main`: The main advertiser email address.
        - `additional`: Optional list of additional advertiser email addresses.
        """
        self.main_advertiser_email = main
        if additional is not None:
            self.advertiser_emails = additional

    def with_agency_emails(self, emails: List[str]):
        """
        Set the agency email addresses.

        **Parameters**:
        - `emails`: List of agency email addresses.
        """
        self.agency_emails = emails

    def with_observer_emails(self, emails: List[str]):
        """
        Set the observer email addresses.

        **Parameters**:
        - `emails`: List of observer email addresses.
        """
        self.observer_emails = emails

    def with_matching_id_format(self, matching_id: MatchingId):
        """
        Set the matching ID format.
        This is required when creating **new** Lookalike Media DCRs only.
        When creating from an existing DCR, the existing matching ID will be used.

        **Parameters**:
        - `matching_id`: The type of matching ID to use.
        """
        self.matching_id = matching_id

    def from_existing(self, lmdcr_id: str, keychain: Keychain):
        """
        Construct a new Lookalike Media DCR from an existing Lookalike Media DCR with the given ID.

        **Parameters**:
        - `lmdcr_id`: The ID of the existing Lookalike Media DCR.
        - `keychain`: The keychain to use to provision datasets from the old Lookalike Media DCR to the new Lookalike Media DCR.
        """
        self.existing = True
        self.lmdcr_id = lmdcr_id
        self.keychain = keychain

    def build_and_publish(self) -> LookalikeMediaDcr:
        """
        Build and publish the Lookalike Media DCR.
        """
        self._check_required_fields_provided()

        if self.existing:
            # Get existing LMDCR (high level from the enclave).
            driver_attestation_hash = self.client._get_lmdcr_driver_attestation_hash(
                self.lmdcr_id
            )
            driver_enclave_spec = self.client._get_enclave_spec_from_hash(
                driver_attestation_hash
            )
            if driver_enclave_spec == None:
                raise Exception(
                    f"Failed to find driver for data room {driver_enclave_spec}"
                )
            session = create_session_from_driver_spec(self.client, driver_enclave_spec)
            existing_lmdcr = session.retrieve_data_room(self.lmdcr_id)
            high_level_representation = json.loads(
                existing_lmdcr.highLevelRepresentation.decode()
            )
            existing = ExistingLookalikeMediaDcr(self.lmdcr_id, driver_enclave_spec)
            return LookalikeMediaDcr(self.client, high_level_representation, existing)
        else:
            id = f"Lookalike DCR {DataRoomBuilder._generate_id()}"
            root_cert_pem = self.client.decentriq_ca_root_certificate.decode()
            (driver_spec, python_spec) = self._get_lmdcr_enclave_specs(self.client)
            (
                matching_id_format,
                matching_id_hashing_algorithm,
            ) = MATCHING_ID_INTERNAL_LOOKUP[self.matching_id]
            # The publisher and advertiser email lists need to contain all emails including the main one.
            publisher_emails = [self.main_publisher_email] + self.publisher_emails
            advertiser_emails = [self.main_advertiser_email] + self.advertiser_emails
            lmdcr = {
                "v3": {
                    "v0": {
                        "id": id,
                        "name": self.name,
                        "mainPublisherEmail": self.main_publisher_email,
                        "mainAdvertiserEmail": self.main_advertiser_email,
                        "matchingIdFormat": matching_id_format,
                        "hashMatchingIdWith": matching_id_hashing_algorithm,
                        "authenticationRootCertificatePem": root_cert_pem,
                        "driverEnclaveSpecification": {
                            "attestationProtoBase64": driver_spec.attestationProtoBase64,
                            "id": driver_spec.id,
                            "workerProtocol": driver_spec.workerProtocol,
                        },
                        "pythonEnclaveSpecification": {
                            "attestationProtoBase64": python_spec.attestationProtoBase64,
                            "id": python_spec.id,
                            "workerProtocol": python_spec.workerProtocol,
                        },
                        "enableAuditLogRetrieval": self.audit_log_retrieval,
                        "enableDevComputations": self.dev_computations,
                        "advertiserEmails": advertiser_emails,
                        "agencyEmails": self.agency_emails,
                        "observerEmails": self.observer_emails,
                        "publisherEmails": publisher_emails,
                    }
                }
            }
            return LookalikeMediaDcr(self.client, lmdcr)

    # Check all necessary options have been provided.
    def _check_required_fields_provided(self):
        # Only perform check when creating new LMDCRs.
        # If creating from existing, these options will be copied across from the existing LMDCR.
        if not self.from_existing:
            if self.name is None:
                raise Exception("A name must be provided")
            elif self.matching_id is None:
                raise Exception("A matching ID must be provided")
            elif self.main_publisher_email is None:
                raise Exception("A publisher email address must be provided")

    @staticmethod
    def _get_lmdcr_enclave_specs(
        client: Client,
    ) -> (compiler.EnclaveSpecification, compiler.EnclaveSpecification):
        enclave_specs = get_latest_enclave_specs_as_dictionary(client)
        driver_spec = {}
        python_spec = {}
        for spec_id, spec in enclave_specs.items():
            spec_payload = {
                "attestationProtoBase64": base64.b64encode(
                    serialize_length_delimited(spec["proto"])
                ).decode(),
                "id": spec_id,
                "workerProtocol": spec["workerProtocols"][0],
            }
            if "decentriq.driver" in spec_id:
                driver_spec = compiler.EnclaveSpecification.parse_obj(spec_payload)
            elif "decentriq.python-ml-worker" in spec_id:
                python_spec = compiler.EnclaveSpecification.parse_obj(spec_payload)
        return (driver_spec, python_spec)
